package com.cyara.apisamples;

import java.io.IOException;
import java.time.format.DateTimeFormatter;
import java.util.ArrayList;
import java.util.Collections;

import javax.json.JsonArray;
import javax.json.JsonObject;

/**
 * For a given timeframe, will grab up to 1000 Pulse calls and
 * produce results showing you every unique prompt, and
 * all the different threshold time results for each prompt.
 * This allows you to adjust threshold times accordingly, great
 * when building out your Pulse library. This program works best
 * when all your prompts are blockified without duplicates.
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.1
 */
public class AdjustThresholds {
	public static void main(String[] argv) throws IOException {
		ApiFunctionalities connInfo = new ApiFunctionalities();

		JsonObject apiObject = connInfo.getTestResultsFeed("pulse", "fromDate","2020-04-01T12:00:00-05:00");
		JsonArray callsArr = apiObject.getJsonArray("results");

		ArrayList<String> expectToHears = new ArrayList<>();
		ArrayList<String> responseTimes = new ArrayList<>();
		ArrayList<String> outcomes = new ArrayList<>();

		/*
		 * Loop through every Pulse call returned by
		 * the API and capture threshold times and
		 * results for every unique prompt.
		 */
		for (int x = 0; x < callsArr.size(); x++) {
			JsonObject call = callsArr.getJsonObject(x);
			JsonObject testStepResultList = call.getJsonObject("testStepResultList");
			JsonArray callStepResult = testStepResultList.getJsonArray("callStepResult");

			//Loop thru all the steps for this call
			for (int stepCnt = 0; stepCnt < callStepResult.size(); stepCnt++) {
				JsonObject stepDetails = callStepResult.getJsonObject(stepCnt);
				JsonObject step = stepDetails.getJsonObject("step");
				JsonObject expect = step.getJsonObject("expect");
				String ethText = expect.getString("text");

				JsonObject stepResult = stepDetails.getJsonObject("stepResult");
				String stepResultDetail = stepResult.getString("detail");

				JsonObject thresholds = step.getJsonObject("threshold");
				String minorRaw = thresholds.getString("minor");
				minorRaw = minorRaw.substring(minorRaw.lastIndexOf(":")+1, minorRaw.length());
				String majorRaw = thresholds.getString("major");
				majorRaw = majorRaw.substring(majorRaw.lastIndexOf(":")+1, majorRaw.length());
				String thresholdFormatted = Integer.parseInt(minorRaw) + "/" + Integer.parseInt(majorRaw);

				//Check if this step has the result detail for breached threshold
				if (stepResultDetail.contains("exceeded")) {
					thresholdFormatted = thresholdFormatted + " BREACHED";
				}
				else {
					thresholdFormatted = thresholdFormatted + " GOOD";
				}

				//If this step is a prompt, process it
				if (ethText.length() > 0) {
					String rawResponseTime = stepDetails.getString("responseTime");
					double time = -1;

					DateTimeFormatter inTime = DateTimeFormatter.ofPattern("HH:mm:ss.SSSSSSS");
					DateTimeFormatter outTime = DateTimeFormatter.ofPattern("s.SSS");
					try{
						time = Double.parseDouble(outTime.format(inTime.parse(rawResponseTime)));
					}catch (Exception e) {
						//e.printStackTrace();
					}

					//Need to check there's actually a time because of optional steps
					if (time > 0) {

						//If this is is a known text, add it to previous index
						if (expectToHears.contains(ethText)) {
							int textLocation = expectToHears.indexOf(ethText);
							responseTimes.set(textLocation, responseTimes.get(textLocation) + time + "~");

							if (thresholdFormatted.contains("BREACHED")) {
								outcomes.set(textLocation, thresholdFormatted);
							}
						}
						else { //If this is an unknown text, add it for the first time
							expectToHears.add(ethText);
							responseTimes.add(time + "~");
							outcomes.add(thresholdFormatted);
						}
					}
				}
			}
		}


		/*
		 * Loop thru all the different prompts found
		 * and print out their overall worst result
		 * and metrics of their threshold times
		 */
		for (int x = 0; x < expectToHears.size(); x++) {
			System.out.println(outcomes.get(x));
			System.out.println(expectToHears.get(x));
			String[] resArr = responseTimes.get(x).split("~");
			ArrayList<Double> resList = new ArrayList<>();

			//Loop thru all the response times and add to arraylist as Double
			for (int res = 0; res < resArr.length; res++) {
				try {
					resList.add(Double.parseDouble(resArr[res]));
				}
				catch (NumberFormatException ex) {

				}
			}

			String outcome = outcomes.get(x);

			//Store this prompt's minor and major thresholds
			int minorThreshold = Integer.parseInt(outcome.substring(0, outcome.indexOf("/")));
			int majorThreshold = Integer.parseInt(outcome.substring(outcome.indexOf("/")+1, outcome.indexOf(" ")));

			Collections.sort(resList);

			//Print the highest and lowest threshold time result, and step usages
			System.out.println("Shortest: " + resList.get(0));
			System.out.println("Longest " + resList.get(resList.size()-1));
			System.out.println("Usages: " + resList.size());


			//Loop to count up all minor and major threshold breaches
			int minorBreaches = 0;
			int majorBreaches = 0;
			for (int n = 0; n < resList.size(); n++) {
				double responseTime = resList.get(n);
				if (responseTime >= majorThreshold) {
					majorBreaches++;
				}
				else if (responseTime >= minorThreshold) {
					minorBreaches++;
				}
			}

			System.out.println("Minor Breaches: " + minorBreaches);
			System.out.println("Major Breaches: " + majorBreaches);

			//Print out all the threshold time results for this prompt
			for (int n = 0; n < resList.size(); n++) {
				System.out.print(resList.get(n));
				if (n < resList.size()-1) {
					System.out.print(" + ");
				}
			}

			System.out.println("\n");

		}

	}

}
