package com.cyara.apisamples;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.io.StringReader;
import java.net.URL;
import java.nio.file.Files;
import java.util.ArrayList;

import javax.json.Json;
import javax.json.JsonObject;
import javax.json.JsonReader;
import javax.net.ssl.HttpsURLConnection;
import javax.sound.sampled.AudioFileFormat.Type;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.spi.AudioFileWriter;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;

import org.w3c.dom.Document;
import org.xml.sax.InputSource;

/**
 * The heart and workhorse of all the API samples, used
 * in every sample. Contains all of the HTTPS requests,
 * the API calls using those requests, the retrieves the
 * credentials to make those API calls. Also contains some
 * basic, reused functions throughout this sample suite. If
 * there's anything you should take and use in your own
 * builds, it's this.
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.6
 */
public class ApiFunctionalities extends AudioFileWriter {
	private String baseUrl = "";
	private int accountNum = -1;
	private String keyToken = "";


	/**
	 * The constructor to create the object.
	 * Gets all your credentials ready to go.
	 */
	public ApiFunctionalities() {

		ArrayList<String> credentials = new ArrayList<String>();
		String fileLine = "";
		//Read in the config information from our credentials file
		try (BufferedReader br = new BufferedReader(new FileReader("credentials.txt"))) {
			while ((fileLine = br.readLine()) != null) {
				credentials.add(fileLine);
			}
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}

		//The base64 encoded string and password, used to create the 3.0 token
		keyToken = credentials.get(0);

		//Which portal we're using.
		baseUrl = credentials.get(1);

		//The account number in the portal we're referencing
		accountNum = Integer.parseInt(credentials.get(2));

	}


	/**
	 * Gets the Portal base Url the boilerplate is using.
	 * The boilerplate originally grabbed it from credentials.txt.
	 * Will be something like https://www.cyaraportal.us/
	 * 
	 * @return The Url for the portal we're using
	 * @since 1.0
	 */
	public String getBaseUrl() {
		return baseUrl;
	}


	/**
	 * Gets the account number the boilerplate is using in
	 * all the API calls. The boilerplate originally grabbed
	 * it from credentials.txt.
	 * 
	 * @return The account number we're making all our APIs against
	 * @since 1.0
	 */
	public int getAccountNum() {
		return accountNum;
	}


	/**
	 * Sets the account number the boilerplate is using in
	 * all the API calls.
	 * 
	 * @param accountNum The account number to change to
	 * @since 1.3
	 */
	public void setAccountNum(int accountNum) {
		this.accountNum = accountNum;
	}


	/**
	 * Convert a String object with Json format into
	 * a JsonObject
	 * 
	 * @param jsonStr The Json String to convert into Json
	 * @return The JsonObject representation of the input String
	 * @since 1.0
	 */
	JsonObject convertAPIOutputToJson(String jsonStr) {
		JsonReader apiReader = Json.createReader(new StringReader(jsonStr));
		JsonObject apiObject = apiReader.readObject();
		apiReader.close();

		return apiObject;
	}


	/**
	 * Convert a String object with XML format into
	 * an XMLDocument
	 * 
	 * @param xmlStr The XML String to convert into XML
	 * @return The XMLDocument representation of the input String
	 * @since 1.0
	 */
	Document convertStringToDocument(String xmlStr) {

		//Boilerplate to convert a string to an XML Document
		DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
		DocumentBuilder builder;
		try
		{
			builder = factory.newDocumentBuilder();
			return builder.parse( new InputSource( new StringReader( xmlStr ) ) );
		} catch (Exception e) {
			e.printStackTrace();
			System.exit(1);
		}

		return null;
	}


	/**
	 * Make an HTTPS GET call using Authorization and Json.
	 * <p>
	 * Used for 3.0 APIs
	 * 
	 * @param apiCall The API endpoint to hit
	 * @return The String of the data sent back from the call
	 * @since 1.0
	 */
	String makeApiGet(String apiCall) {
		String resultLine = "";

		StringBuilder apiSB = new StringBuilder();

		try {
			URL apiURL = new URL(apiCall);
			HttpsURLConnection apiCon = (HttpsURLConnection) apiURL.openConnection();
			apiCon.setRequestProperty("Authorization", keyToken);
			apiCon.setRequestProperty("Accept", "application/json");
			InputStream apiInStream = apiCon.getInputStream();
			InputStreamReader apiInStreamReader = new InputStreamReader(apiInStream);
			BufferedReader apiBufRead = new BufferedReader(apiInStreamReader);

			while ((resultLine = apiBufRead.readLine()) != null) {
				apiSB.append(resultLine);
			}

			apiBufRead.close();
		}
		catch (IOException ex) {
			ex.printStackTrace();
		}

		return apiSB.toString();
	}


	/**
	 * Make an HTTPS GET call using Authorization and Json
	 * to retrieve and download a .wav file
	 * <p>
	 * Used for 3.0 APIs
	 * 
	 * @param apiCall The API endpoint to hit
	 * @return The downloaded audio file path
	 * @since 1.0
	 */
	String makeApiGetWav(String apiCall) {

		String audioFilePath = "";

		try {
			URL apiURL = new URL(apiCall);
			HttpsURLConnection apiCon = (HttpsURLConnection) apiURL.openConnection();
			apiCon.setRequestProperty("Authorization", keyToken);
			apiCon.setRequestProperty("Accept", "*/*");
			apiCon.setRequestProperty("Accept-Encoding", "gzip, deflate, br");
			apiCon.setRequestProperty("Connection", "keep-alive");
			InputStream apiInStream = apiCon.getInputStream();

			String disposition = apiCon.getHeaderField("Content-Disposition");

			String fileName = disposition.substring(disposition.indexOf("=")+1, disposition.length());

			audioFilePath = fileName.replace("_", "");
			audioFilePath = audioFilePath.substring(audioFilePath.indexOf("Result")+6, audioFilePath.length());

			// opens an output stream for writing file
			FileOutputStream outputStream = new FileOutputStream(audioFilePath);

			byte[] buffer = new byte[1024];
			int bytesRead = -1;

			while ((bytesRead = apiInStream.read(buffer)) != -1) {
				outputStream.write(buffer, 0, bytesRead);
			}

			outputStream.close();
			apiInStream.close();
		}
		catch (IOException ex) {
			ex.printStackTrace();
		}

		return audioFilePath;
	}


	/**
	 * Make an HTTPS PUT/POST/DELETE call using Authorization and Json.
	 * <p>
	 * Used for 3.0 APIs
	 * 
	 * @param httpMethod The type of call made, either PUT, POST, or DELETE
	 * @param apiCall The API endpoint to hit
	 * @param body Any json passed into the API request
	 * @param contentType The type of data we're passing into the API
	 * @return The JsonObject of the data sent back from the call
	 * @since 1.0
	 */
	JsonObject makeApiPutPostDel(String httpMethod, String apiCall, String body, String contentType) {
		String resultLine = "";

		StringBuilder apiSB = new StringBuilder();

		try {
			URL apiURL = new URL(apiCall);
			HttpsURLConnection apiCon = (HttpsURLConnection) apiURL.openConnection();
			apiCon.setRequestMethod(httpMethod);
			apiCon.setRequestProperty("Authorization", keyToken);
			apiCon.setRequestProperty("Content-Type", contentType);
			apiCon.setRequestProperty("Accept", "application/json");
			apiCon.setDoOutput(true);
			OutputStreamWriter apiWriter = new OutputStreamWriter(apiCon.getOutputStream(), "UTF-8");
			apiWriter.write(body);
			apiWriter.close();

			BufferedReader apiBufRead = new BufferedReader(new InputStreamReader(apiCon.getInputStream()));

			while ((resultLine = apiBufRead.readLine()) != null) {
				apiSB.append(resultLine);
			}

			apiBufRead.close();
		}
		catch (IOException ex) {
			ex.printStackTrace();
		}

		return convertAPIOutputToJson(apiSB.toString());
	}


	/**
	 * Make an HTTPS PUT/POST/DELETE call using Authorization and Json.
	 * Used only for datasets because that API doesn't return anything.
	 * <p>
	 * Used for 3.0 APIs
	 * 
	 * @param httpMethod The type of call made, either PUT, POST, or DELETE
	 * @param apiCall The API endpoint to hit
	 * @param body Any json passed into the API request
	 * @param contentType The type of data we're passing into the API
	 * @return The HTTP Status of the call
	 * @since 1.1
	 */
	int datasetPutPostDel(String httpMethod, String apiCall, String body, String contentType) {
		int status = -1;
		String resultLine = "";

		StringBuilder apiSB = new StringBuilder();

		try {
			URL apiURL = new URL(apiCall);
			HttpsURLConnection apiCon = (HttpsURLConnection) apiURL.openConnection();
			apiCon.setRequestMethod(httpMethod);
			apiCon.setRequestProperty("Authorization", keyToken);
			apiCon.setRequestProperty("Content-Type", contentType);
			apiCon.setRequestProperty("Accept", "application/json");
			apiCon.setDoOutput(true);
			OutputStreamWriter apiWriter = new OutputStreamWriter(apiCon.getOutputStream(), "UTF-8");
			apiWriter.write(body);
			apiWriter.close();

			BufferedReader apiBufRead = new BufferedReader(new InputStreamReader(apiCon.getInputStream()));

			while ((resultLine = apiBufRead.readLine()) != null) {
				apiSB.append(resultLine);
			}

			status = apiCon.getResponseCode();

			apiBufRead.close();
		}
		catch (IOException ex) {
			ex.printStackTrace();
		}

		return status;
	}


	/**
	 * Makes an API call to list up to 1000 accounts the
	 * user is attached to.
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Accounts/Accounts_GetAccounts
	 * 
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject listAccounts() {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts?sortField=accountId&sortAsc=true&pageSize=1000"));
	}


	/**
	 * Makes an API call to return the account settings
	 * for the supplied account.
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Accounts/Accounts_GetAccount
	 * 
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject getAccount() {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum));
	}


	/**
	 * Makes an API call to list up to 1000 CX Models located
	 * in the specified account
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/CXModels/CXModels_GetModelsForAccount
	 * 
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject listCXModels() {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/models?sortField=Name&sortAsc=true&pageSize=1000"));
	}


	/**
	 * Makes an API call to export the specified CX Model into
	 * CX Model Json
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/CXModels/CXModels_GetModelById
	 * 
	 * @param cxModelId The ID of the CX Model to export
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject exportCXModel(int cxModelId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/models/" + cxModelId));
	}


	/**
	 * Makes an API call to import CX Model Json into the
	 * specified account, and overwrite or create a new
	 * CX Model.
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/CXModels/CXModels_CreateModel
	 * 
	 * @param cxModelJson The json being imported to create the CX Model
	 * @param createMode The type of import you want to do. Can
	 * be Default UniqueName OverrideExisting
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject importCXModel(String cxModelJson, String createMode) {
		return makeApiPutPostDel("POST", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/models?createMode=" + createMode, cxModelJson, "application/json");
	}


	/**
	 * Makes an API call to list up to 1000 Dashboards located
	 * in the specified account
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Dashboards/Dashboards_GetDashboards
	 * 
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject listDashboards() {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/dashboards?sortField=Name&sortAsc=true&pageSize=1000"));
	}


	/**
	 * Retrieves all the results and data used to create a
	 * Cyara Executive Dashboard up to the specified number
	 * of hours ago for the desired dashboard
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Dashboards/Dashboards_GetDashboardFeed
	 * 
	 * @param dashboardId The ID of the dashboard whose feed we want
	 * @param hours The amount of hours ago we grab data up to
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject getDashboardFeed(int dashboardId, int hours) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/dashboards/" + dashboardId + "/feed?maxHours=" + hours));
	}


	/**
	 * Makes an API call to retrieve up to 1000 tickers
	 * that currently exist and are displayed for a given dashboard
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/DashboardTickers/DashboardTickers_GetDashboardTickers
	 * 
	 * @param dashboardId The ID of the dashboard whose feed we want
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject getDashboardTickers(int dashboardId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/dashboards/" + dashboardId + "/tickers?sortField=DateExpires&sortAsc=true&pageSize=1000"));
	}


	/**
	 * Makes an API call to list up to 1000 datasets located
	 * in the specified account
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Datasets/Datasets_GetDatasetsForAccount
	 * 
	 * @param dsSearchTerm The substring DS name to filter by
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject listDatasets(String dsSearchTerm) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/datasets?searchTerm=" + dsSearchTerm + "&pageSize=1000"));
	}


	/**
	 * Makes an API call to return a specified dataset,
	 * including the headers and number of scenarios
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Datasets/Datasets_GetDatasetsForAccount
	 * 
	 * @param datasetId The ID of the dataset to retrieve
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject getDataset(int datasetId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/datasets/" + datasetId));
	}


	/**
	 * Makes an API call to return up to 1000 scenarios (rows)
	 * for a specified dataset,
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Datasets/Datasets_GetDataRows
	 * 
	 * @param datasetId The ID of the dataset to retrieve rows for
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject getDatasetRows(int datasetId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/datasets/" + datasetId + "/rows?sortAsc=true&pageSize=1000"));
	}


	/**
	 * Makes an API call to return up to 1000 test cases
	 * currently using the specified dataset
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Datasets/Datasets_DatasetUsage
	 * 
	 * @param datasetId The ID of the dataset to find usages for
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject getDatasetTestCases(int datasetId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/datasets/" + datasetId + "/testcases?sortAsc=true&pageSize=1000"));
	}


	/**
	 * Makes an API call to associate the specified
	 * dataset to the specified test case
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Datasets/Datasets_LinkTestCaseWithDataset
	 * 
	 * @param datasetId The ID of the dataset to link to the test case
	 * @param testCaseId The ID of the test case to associate with the dataset
	 * @return The HTTP Status of the call
	 * @since 1.0
	 */
	int linkDataset(int datasetId, int testCaseId) {
		return datasetPutPostDel("PUT", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/datasets/" + datasetId + "/testcases/" + testCaseId, "{}", "application/json");
	}


	/**
	 * Permanently and irreversibly deletes the single, specified dataset
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Datasets/Datasets_DeleteDatasetFromAccount
	 * 
	 * @param datasetId The ID of the dataset to delete
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject deleteDataset(int datasetId) {
		return makeApiPutPostDel("DELETE", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/datasets/" + datasetId, "{}", "application/json");
	}


	/**
	 * Permanently and irreversibly deletes all the specified datasets
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Datasets/Datasets_DeleteDatasetsFromAccount
	 * 
	 * @param datasetIds The IDs of the datasets to delete in the form of "[146,147]"
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject deleteDatasets(String datasetIds) {
		return makeApiPutPostDel("DELETE", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/datasets", datasetIds, "application/json");
	}


	/**
	 * Takes in a valid .csv file and imports it into the portal,
	 * creating a brand new dataset
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Datasets/Datasets_AddDatasetToAccount
	 * 
	 * @param csvLocation The filepath of your .csv file
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject createDataset(String csvLocation) {
		return filePutPost("POST", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/datasets", csvLocation, "csv");
	}


	/**
	 * Takes in a valid .csv file and imports it into the portal,
	 * updating an existing, specified dataset. Replaces all
	 * the existing content in the dataset with the .csv content
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Datasets/Datasets_ReplaceDataset
	 * 
	 * @param csvLocation The filepath of your .csv file
	 * @param datasetId The ID of the dataset to update
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject updateDataset(String csvLocation, int datasetId) {
		return filePutPost("PUT", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/datasets/" + datasetId + "/rows", csvLocation, "csv");
	}


	/**
	 * Takes in a valid Cyara XML file and imports it into the portal,
	 * creating a new test case or updating an existing one. A new test
	 * case is create if the folder name and test case name combination
	 * doesn't exist. Otherwise, the existing test case is updated.
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/TestCases/TestCases_ImportTestCases
	 * 
	 * @param xmlLocation The filepath of your .xml file
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject importTestCaseFromFile(String xmlLocation) {
		return filePutPost("POST", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/testcases/import", xmlLocation, "xml");
	}


	/**
	 * Takes in a valid Cyara XML String and imports it into the portal,
	 * creating a new test case or updating an existing one. A new test
	 * case is create if the folder name and test case name combination
	 * doesn't exist. Otherwise, the existing test case is updated.
	 * <p>
	 * This call is not yet documented in Swagger
	 * 
	 * @param rawXML The XML String of the test case to import
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject importTestCaseFromString(String rawXML) {
		return makeApiPutPostDel("POST", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/testcases/import", rawXML, "application/xml");
	}


	/**
	 * Makes an HTTP PUT or POST with a file attached as part of the call
	 * 
	 * @param httpMethod The type of call made, either PUT or POST
	 * @param apiCall The API endpoint to hit
	 * @param fileLocation The filepath of the file to attach to the HTTP request
	 * @param fileType The type of file we're passing into the request
	 * @return The JsonObject of the data sent back from the call
	 * @since 1.0
	 */
	JsonObject filePutPost(String httpMethod, String apiCall, String fileLocation, String fileType) {
		File textFile = new File(fileLocation);
		String boundary = Long.toHexString(System.currentTimeMillis()); // Just generate some unique random value.
		String crlf = "\r\n"; // Line separator required by multipart/form-data.
		String charset = "UTF-8";

		StringBuilder apiSB = new StringBuilder();

		try {
			URL apiURL = new URL(apiCall);
			HttpsURLConnection apiCon = (HttpsURLConnection) apiURL.openConnection();
			apiCon.setRequestMethod(httpMethod);
			apiCon.setRequestProperty("Authorization", keyToken);
			apiCon.setRequestProperty("Content-Type", "multipart/form-data; boundary=----" + boundary);
			apiCon.setRequestProperty("Accept", "application/json");
			apiCon.setDoOutput(true);

			OutputStream output = apiCon.getOutputStream();
			PrintWriter apiWriter = new PrintWriter(new OutputStreamWriter(output, charset), true);

			// Send text file.
			apiWriter.append("------" + boundary).append(crlf);
			apiWriter.append("Content-Disposition: form-data; name=\"Choose file\"; filename=\"" + textFile.getName() + "\"").append(crlf);
			apiWriter.append("Content-Type: text/" + fileType + "; charset=UTF-8").append(crlf); // Text file itself must be saved in this charset!
			apiWriter.append(crlf).flush(); 
			Files.copy(textFile.toPath(), output);
			output.flush(); // Important before continuing with writer!
			apiWriter.append(crlf).flush(); // CRLF is important! It indicates end of boundary.

			// End of multipart/form-data.
			apiWriter.append(crlf).flush();
			apiWriter.append("------" + boundary + "--").append(crlf).flush();
			apiWriter.close();

			BufferedReader importBR = new BufferedReader(new InputStreamReader(apiCon.getInputStream()));

			String resultLine = "";
			while ((resultLine = importBR.readLine()) != null) {
				apiSB.append(resultLine);
			}

			importBR.close();
		}
		catch (IOException ex) {
			ex.printStackTrace();
		}

		return convertAPIOutputToJson(apiSB.toString());
	}


	/**
	 * Exports one or more test cases as Cyara XML
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/TestCases/TestCases_GetCyaraXml
	 * 
	 * @param testCaseIds The IDs of all the test cases to export
	 * @param blockOption How to export the blocks the test cases are using IncludeBlocks ExcludeBlocks FlattenBlocks
	 * @return The String Cyara XML of the exported test cases
	 * @since 1.0
	 */
	String exportTestCases(ArrayList<Integer> testCaseIds, String blockOption) { //&testCaseIds=1234&testCaseIds=432
		StringBuilder testCasesParam = new StringBuilder();
		for (int x = 0; x < testCaseIds.size(); x++) {
			testCasesParam.append("&testCaseIds=" + testCaseIds.get(x));
		}
		return makeApiGet(baseUrl + "CyaraWebApi/v3.0/accounts/" + accountNum + "/testcases/cyaraxml?&recursive=false&includeDataDriven=true&includeServices=true&download=false&blockOptions=" + blockOption + "&exportFormat=XML" + testCasesParam.toString());
	}


	/**
	 * Exports a single test case as Cyara XML
	 * 
	 * @param testCaseId The ID of the test case to export
	 * @param blockOption How to export the blocks the test case is using IncludeBlocks ExcludeBlocks FlattenBlocks
	 * @see exportTestCases
	 * @return The String Cyara XML of the exported test case
	 * @since 1.0
	 */
	String exportTestCase(int testCaseId, String blockOption) {
		ArrayList<Integer> testCaseIdList = new ArrayList<>();
		testCaseIdList.add(testCaseId);
		return exportTestCases(testCaseIdList, blockOption);
	}


	/**
	 * Exports one or more blocks as Cyara XML
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Blocks/Blocks_GetCyaraXml
	 * 
	 * @param blockIds The IDs of all the blocks to export
	 * @return The String Cyara XML of the exported blocks
	 * @since 1.0
	 */
	String exportBlocks(ArrayList<Integer> blockIds) {
		StringBuilder blockParam = new StringBuilder();
		for (int x = 0; x < blockIds.size(); x++) {
			blockParam.append("&blockIds=" + blockIds.get(x));
		}
		return makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/blocks/cyaraxml?exportFormat=XML" + blockParam);
	}


	/**
	 * Exports a single block as Cyara XML
	 * 
	 * @param blockId The ID of the block to export
	 * @see exportBlocks
	 * @return The String Cyara XML of the exported block
	 * @since 1.0
	 */
	String exportBlock(int blockId) {
		ArrayList<Integer> blockIdList = new ArrayList<>();
		blockIdList.add(blockId);
		return exportBlocks(blockIdList);
	}


	/**
	 * Get the basic details for the supplied test case.
	 * <p>
	 * Does not export as Cyara XML.
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/TestCases/TestCases_GetTestCase
	 * 
	 * @param testCaseId The ID of the test case to get information on
	 * @see exportTestCase
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject getTestCaseBasics(int testCaseId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/testcases/" + testCaseId));
	}


	/**
	 * Retrieve all the test cases using the specified block
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Blocks/Blocks_GetLinkedTestCaseDetailsByBlockId
	 * 
	 * @param blockId The ID of the block to check usage on
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject getLinkedTestCases(int blockId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/blocks/" + blockId + "/testcases"));
	}


	/**
	 * Retrieve up to 1000 calls occurring from the specified plan type
	 * since the specified date or result ID. The calls will be returned
	 * with every step of every call and every detail of every step of every call.
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/TestResults/TestResults_GetTestResultsFeed
	 * 
	 * @param planType The plan type to retrieve calls for
	 * @param fromType The choice to use a date or result ID to grab results from
	 * @param fromRef The earliest date or result ID to grab calls from. Can be
	 * fromDate or fromTestResultId. If using fromDate, formats can be
	 * 2020-04-01T12:00:00-05:00 or 2020-02-14T15:25:56Z or 2020-05-15T05:15:27.048638600Z
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject getTestResultsFeed(String planType, String fromType, String fromRef) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/Voice/testresults/feed?planType=" + planType + "&" + fromType + "=" + fromRef + "&stepResults=true&maxResults=1000"));
	}


	/**
	 * Immediately starts the specified campaign
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Campaigns/Campaigns_CampaignExecute
	 * 
	 * @param campaignId The ID of the campaign to run
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject executeCampaign(int campaignId) {
		return makeApiPutPostDel("POST", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/campaigns/" + campaignId + "/execute", "", "application/json");
	}


	/**
	 * Gets the status of a campaign run
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/CampaignRuns/CampaignRuns_GetCampaignRun
	 * 
	 * @param campaignId The ID of the campaign we're running
	 * @param campaignRunId The GUID for the campaign execution
	 * @return The JsonObject of what the API returned
	 * @since 1.5
	 */
	JsonObject getCampaignRunByCampaignRunId(int campaignId, String campaignRunId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/campaigns/" + campaignId + "/runs/" + campaignRunId));
	}

	/**
	 * Gets the results of a campaign run
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Reports/Reports_GetCampaignRunTestResults
	 * 
	 * @param campaignId The ID of the campaign whose results we want
	 * @param runId The ID for the campaign run's report
	 * @return The JsonObject of what the API returned
	 * @since 1.5
	 */
	JsonObject getCampaignRunTestStepResults(int campaignId, int runId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/report/campaigns/" + campaignId + "/runs/" + runId + "/results?stepResults=true&pageSize=1000"));
	}


	/**
	 * Gets the run history of a campaign
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Campaigns/Campaigns_GetCampaignRunHistory
	 * 
	 * @param campaignId The ID of the campaign we're running
	 * @return The JsonObject of what the API returned
	 * @since 1.5
	 */
	JsonObject getCampaignRunHistory(int campaignId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/campaigns/" + campaignId + "/runs?sortAsc=false&pageSize=1000"));
	}


	/**
	 * Retrieve the results of a test case executed by the API
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/TestResults/TestResults_GetTestResultByTicketId
	 * 
	 * @param ticketId The ticket ID of the test case run used by the portal
	 * @see executeTestCase
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject getTestCaseResultByTicketId(String ticketId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/testresults/" + ticketId));
	}


	/**
	 * Runs the test case as a single validation run
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/TestCases/TestCases_ExecuteTestCase
	 * 
	 * @param testCaseId The ID of the test case to run
	 * @see getTestCaseResultByTicketId
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject executeTestCase(int testCaseId) {
		return makeApiPutPostDel("POST", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/testcases/" + testCaseId + "/execute", "", "application/json");
	}


	/**
	 * List up to 1000 blocks on the account
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Blocks/Blocks_GetBlockListByAccount
	 * 
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject listBlocks() {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/blocks?channel=Voice&pageSize=1000"));
	}


	/**
	 * List up to 1000 test cases on the account
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/TestCases/TestCases_GetTestCases
	 * 
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject listTestCases() {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/testcases?channel=Voice&pageSize=1000"));
	}


	/**
	 * Returns all the plan details for the specified plan,
	 * including the expiration and number of ports
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Plans/Plans_GetPlan
	 * 
	 * @param planId The ID of the plan to retrieve details for
	 * @return The JsonObject of what the API returned
	 * @since 1.0
	 */
	JsonObject getPlanDetails(int planId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/plans/" + planId));
	}


	/**
	 * Creates a campaign using the 3.0 API
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/Campaigns/Campaigns_CreateCampaign
	 * 
	 * @param campaignJson The configuration of the campaign we're building
	 * @return The JsonObject of what the API returned
	 * @since 1.1
	 */
	JsonObject createCampaign(String campaignJson) {
		return makeApiPutPostDel("POST", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/campaigns", campaignJson, "application/json");
	}


	/**
	 * Makes an API call to recordings
	 * available per step for a given
	 * test case result.
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/TestResults/TestResults_GetCallRecordingList
	 * 
	 * @param testResultId The ID of the voice test case result
	 * @return The JsonObject of what the API returned
	 * @since 1.3
	 */
	JsonObject getStepRecordingAvailabilities(int testResultId) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/voice/testresults/" + testResultId + "/steps/recordingstate"));
	}


	/**
	 * Makes an API call to get the
	 * recording for a given step.
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/TestResults/TestResults_GetFullCallRecording
	 * 
	 * @param testResultId The ID of the voice test case result
	 * @param step The step whose recording to retrieve
	 * @return The downloaded audio file path
	 * @since 1.3
	 */
	String getRecording(int testResultId, int step) {
		String stepInfo = "";
		if (step > 0) {
			stepInfo = "?stepNo=" + step;
		}
		return makeApiGetWav(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/voice/recording/" + testResultId + stepInfo);
	}


	/**
	 * Makes an API call to get the step
	 * transcription for any step in a voice
	 * call.
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/TranscriptionProxy/TranscriptionProxy_GetTranscription
	 * 
	 * @param testResultId The ID of the voice test case result
	 * @param step The step whose transcription to retrieve
	 * @return The JsonObject of what the API returned
	 * @since 1.3
	 */
	JsonObject getStepTranscription(int testResultId, int step) {
		return convertAPIOutputToJson(makeApiGet(baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/transcription/testresults/" + testResultId + "/steps/" + step));
	}


	/**
	 * Makes an API call to generate the step
	 * transcription for any step in a voice
	 * call.
	 * <p>
	 * https://www.cyaraportal.us/cyarawebapi/swagger/index.html?url=/cyarawebapi/swagger/v3/swagger.json#/TranscriptionProxy/TranscriptionProxy_Transcribe
	 * 
	 * @param testResultId The ID of the voice test case result
	 * @param step The step whose transcription to retrieve
	 * @return The JsonObject of what the API returned
	 * @since 1.3
	 */
	JsonObject generateStepTranscription(int testResultId, int step) {
		return  makeApiPutPostDel("POST", baseUrl + "cyarawebapi/v3.0/accounts/" + accountNum + "/transcription/testresults/" + testResultId + "/steps/" + step + "?forceTranscription=false",
				"{speakerName: \"English (USA), en-US, v2\", speakerTechnology: \"Google\", speakerTag: \"en-US-G\"}", "application/json");
	}


	@Override
	public Type[] getAudioFileTypes() {
		return null;
	}


	@Override
	public Type[] getAudioFileTypes(AudioInputStream stream) {
		return null;
	}


	@Override
	public int write(AudioInputStream stream, Type fileType, OutputStream out) throws IOException {
		return 0;
	}


	@Override
	public int write(AudioInputStream stream, Type fileType, File out) throws IOException {
		return 0;
	}

}