package com.cyara.apisamples;

import java.time.Instant;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;

import javax.json.JsonArray;
import javax.json.JsonObject;
import javax.json.stream.JsonParsingException;

import com.cyara.utilities.Utils;

/**
 * A program to automate the creation and usage of a dataset.
 * Once the dataset is created, it's attached to a test case,
 * which is then used in creating a campaign. That campaign is then run.
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.2
 */
public class DatasetEndToEnd {

	//Change to your test case ID
	static int testCaseId = -1;

	//Change to your Velocity Plan ID
	static int velocityPlanId = -1;

	//Change to your dataset csv file path on your local machine
	static String csvFilePath = "CHANGE_ME.csv";

	public static void main(String[] argv) {
		ApiFunctionalities connInfo = new ApiFunctionalities();

		int datasetId = checkForExistingDS(connInfo, getLocalDSName());

		/*
		 * There is no existing dataset with the same name as our file.
		 * We must create a new dataset.
		 */
		if (datasetId == -1) {
			System.out.println("No existing dataset matched, creating a new one");

			//Create the dataset from your csv file
			JsonObject datasetCreation = connInfo.createDataset(csvFilePath);

			//Get the dataset ID from the dataset we just created
			datasetId = datasetCreation.getInt("datasetId");
			System.out.println("Dataset created with ID: " + datasetId);
		}

		//Attach the newly-created dataset to our test case
		int linkStatus = connInfo.linkDataset(datasetId, testCaseId);

		if (linkStatus == 200) {

			System.out.println("Dataset successfully linked to test case");

			//Run test case and check if it's Success
			if (validateTestCaseAndGetResults(connInfo) == 0) {

				//Run the campaign and keeps checking until it's done
				int campaignId = runCampaignAndCheckForCompletion(connInfo);

				//Pull the most recent run from the historical runs
				int runId = getMostRecentRunId(connInfo, campaignId);

				System.out.println("The runId for the finished campaign execution: " + runId);

				//Get the results of the run, and process
				processCampaignRun(connInfo, campaignId, runId);
			}
			else {
				System.err.println("ERROR: Test Case Validation failed for TC ID: " + testCaseId +
						"The campaign was not created and not run. Please remedy TC first then rerun");
			}
		}
	}


	/**
	 * Grabs the campaign run's test step results, getting
	 * every call, every step of every call, and every detail
	 * of every step of every call. The results of each step
	 * are then parsed and printed out.
	 * 
	 * @param connInfo The controller to the API credentials and methods
	 * @param campaignId The ID of the campaign whose history we want
	 * @param runId The id for the campaign run's report
	 * @since 1.2
	 */
	private static void processCampaignRun(ApiFunctionalities connInfo, int campaignId, int runId) {
		boolean morePages = true;
		String nextPage = "";

		JsonObject campaignRunTestStepResults = connInfo.getCampaignRunTestStepResults(campaignId, runId);

		//Loop through all the calls and their step results
		while (morePages) {
			if (campaignRunTestStepResults.containsKey("nextPage")) {
				if (campaignRunTestStepResults.isNull("nextPage")) {
					morePages = false;
				}
				else {
					nextPage = campaignRunTestStepResults.getString("nextPage");
				}

				JsonArray callResultsArr = campaignRunTestStepResults.getJsonArray("results");
				for (int callIndex = 0; callIndex < callResultsArr.size(); callIndex++) {
					//Get a single call from the campaign
					JsonObject callResult = callResultsArr.getJsonObject(callIndex);

					//Get the breakdown of steps for this call
					JsonObject testStepResultList = callResult.getJsonObject("testStepResultList");

					//Get the Step 1 and beyond results for this call
					JsonArray callStepResultsArr = testStepResultList.getJsonArray("callStepResult");

					//Loop thru each step of this call and prints its result
					for (int stepIndex = 0; stepIndex < callStepResultsArr.size(); stepIndex++) {
						JsonObject callStepResult = callStepResultsArr.getJsonObject(stepIndex);
						JsonObject stepResult = callStepResult.getJsonObject("stepResult");
						System.out.println(stepResult.getString("result"));
					}
				}

			}
			else { //nextPage doesn't even exist, something is totally wrong
				morePages = false;
			}

			if (morePages) { //After processing that call, do we still have pages? Then make the next call here
				campaignRunTestStepResults = connInfo.convertAPIOutputToJson(connInfo.makeApiGet(nextPage));
			}
		}
	}


	/**
	 * Grabs a campaign's run history, and pulls its most recent
	 * run, grabbing the runId.
	 * 
	 * @param connInfo The controller to the API credentials and methods
	 * @param campaignId The ID of the campaign whose history we want
	 * @return The most recent runId for this campaign
	 * @since 1.2
	 */
	private static int getMostRecentRunId(ApiFunctionalities connInfo, int campaignId) {
		JsonObject campaignRunHistory = connInfo.getCampaignRunHistory(campaignId);

		JsonArray results = campaignRunHistory.getJsonArray("results");

		JsonObject mostRecentRun = results.getJsonObject(0);

		return mostRecentRun.getInt("runId");
	}


	/**
	 * Execute (run) the campaign and check every so often if
	 * it's finished running.
	 * 
	 * @param connInfo The controller to the API credentials and methods
	 * @return The newly-created campaign's ID
	 * @since 1.2
	 */
	private static int runCampaignAndCheckForCompletion(ApiFunctionalities connInfo) {
		//Create the campaign
		int campaignId = campaignCreation(connInfo).getInt("campaignId");

		System.out.println("Campaign created with ID: " + campaignId);

		//Run our newly-created campaign
		JsonObject campaignExecution = connInfo.executeCampaign(campaignId);

		String campaignRunId = campaignExecution.getString("campaignRunId");

		//Loop to check over and over if the campaign is done running
		while (true) {
			JsonObject campaignRun = connInfo.getCampaignRunByCampaignRunId(campaignId, campaignRunId);

			String status = campaignRun.getString("status");

			if (!status.equals("Running") && !status.equals("Aborting")) {
				break;
			}
			else {
				System.out.println("Campaign is still running. Still waiting...");
				Utils.sleepMode(20000);
			}
		}

		return campaignId;
	}


	/**
	 * Validate the test case and check for completion until it's finished.
	 * Once it is, grab the overall test result for the call.
	 * 
	 * @param connInfo The controller to the API credentials and methods
	 * @return The test result. 0 for Success, -1 for other
	 * @since 1.2
	 */
	private static int validateTestCaseAndGetResults(ApiFunctionalities connInfo) {
		//Run the test case
		JsonObject testCaseExecAck = connInfo.executeTestCase(testCaseId);

		//Get the testRunTicketId for the run
		String testRunTicketId = testCaseExecAck.getString("testRunTicketId");

		int validationResult = 0;

		//Loop to keep checking if the call is done
		while (true) {
			try {

				//Get the call result (will throw an error if call isn't done)
				JsonObject testCaseResult = connInfo.getTestCaseResultByTicketId(testRunTicketId);

				JsonObject testResult = testCaseResult.getJsonObject("testResult");
				String result = testResult.getString("result");

				if (!result.equals("Success")) {
					validationResult = -1;
				}
				break;
			}
			catch (JsonParsingException ex) {
				/*
				 * The API returns a 404 while the call is
				 * still running. Because of this, there's
				 * no Json to parse, thus the only way
				 * to determine the call is still running
				 * is this try, catch, and sleep until retry.
				 */
				System.out.println("Test case is still running. Still waiting...");
				Utils.sleepMode(20000);
			}
		}

		return validationResult;
	}


	/**
	 * Takes the full path of our local csv file and parses out
	 * the path and file type. This leaves us with just the file
	 * name, which is what we'll eventually check for in our
	 * existing datasets.
	 * 
	 * @return The file name without the file type and path
	 * @see checkForExistingDS
	 * @since 1.2
	 */
	private static String getLocalDSName() {
		String localDatasetName = "";

		if (csvFilePath.contains("/")) {
			localDatasetName = csvFilePath.substring(csvFilePath.lastIndexOf("/")+1, csvFilePath.lastIndexOf("."));
		}
		else if (csvFilePath.contains("\\")) {
			localDatasetName = csvFilePath.substring(csvFilePath.lastIndexOf("\\")+1, csvFilePath.lastIndexOf("."));
		}
		else {
			localDatasetName = csvFilePath.substring(0, csvFilePath.lastIndexOf("."));
		}

		return localDatasetName.toLowerCase();
	}


	/**
	 * Loop through all existing datasets on the account. Check if
	 * the name of the dataset matches our local file. If we find
	 * one that does, grab and return the ID.
	 * 
	 * @param connInfo The controller to the API credentials and methods
	 * @param localDSName The lowercase local csv file name
	 * @return The matching existing dataset ID, or -1
	 * @since 1.2
	 */
	private static int checkForExistingDS(ApiFunctionalities connInfo, String localDSName) {
		boolean morePages = true;
		String nextPage = "";

		int existingMatchingDSId = -1;

		JsonObject allDatasetsObject = connInfo.listDatasets(localDSName);

		System.out.println("Checking for existing dataset with name (case-insensitive): " + localDSName);

		/*
		 * Loop through all the pages of listed datasets, looking
		 * for an existing dataset with the same name as our file.
		 */
		while (morePages) {
			if (allDatasetsObject.containsKey("nextPage")) {
				if (allDatasetsObject.isNull("nextPage")) {
					morePages = false;
				}
				else {
					nextPage = allDatasetsObject.getString("nextPage");
				}

				JsonArray datasetArr = allDatasetsObject.getJsonArray("results");
				for (int datasetIndex = 0; datasetIndex < datasetArr.size(); datasetIndex++) {
					JsonObject dataset = datasetArr.getJsonObject(datasetIndex);
					String datasetName = dataset.getString("name").toLowerCase();

					if (localDSName.equals(datasetName)) {
						existingMatchingDSId = dataset.getInt("datasetId");
						System.out.println("Found existing matching dataset with ID: " + existingMatchingDSId);
						break;
					}
				}

			}
			else { //nextPage doesn't even exist, something is totally wrong
				morePages = false;
			}

			if (morePages) { //After processing that call, do we still have pages? Then make the next call here
				allDatasetsObject = connInfo.convertAPIOutputToJson(connInfo.makeApiGet(nextPage));
			}
		}

		return existingMatchingDSId;
	}


	/**
	 * Creates a new Velocity campaign that uses our dataset-driven
	 * test case.
	 * 
	 * @param connInfo The controller to the API credentials and methods
	 * @return The JsonObject for the newly-created campaign
	 * @since 1.2
	 */
	private static JsonObject campaignCreation(ApiFunctionalities connInfo) {
		/*
		 * Create a date in the future so the campaign doesn't immediately run
		 * If we create a date in the past, the campaign would run immediately
		 * and break this program.
		 */
		String requestedScheduledRunDate = DateTimeFormatter.ISO_INSTANT.format(Instant.now().plus(1, ChronoUnit.DAYS));

		//The json of the velocity campaign to create that will run our test case
		String campaignBody = "{\n" + 
				" \"campaignSync\": {\n" + 
				"  \"isActive\": false\n" + 
				" },\n" + 
				" \"pauseTimeBetweenRetries\": null,\n" + 
				" \"maximumRetries\": null,\n" + 
				" \"concurrency\": 1,\n" + 
				" \"maximumAps\": 1,\n" + 
				" \"active\": true,\n" + 
				" \"channel\": \"Voice\",\n" + 
				" \"description\": \"This campaign runs a test case with a dataset. All created by the API\",\n" + 
				" \"name\": \"Dataset API Campaign\",\n" + 
				" \"testCaseDistributionProfile\": \"RoundRobin\",\n" + 
				" \"planType\": \"Velocity\",\n" + 
				" \"planId\": " + velocityPlanId + ",\n" + 
				" \"requestedRunDate\": \"" + requestedScheduledRunDate + "\",\n" + 
				" \"scheduledRunDate\": \"" + requestedScheduledRunDate + "\",\n" + 
				" \"testCases\": [\n" + 
				"  {\n" + 
				"   \"testCaseId\": " + testCaseId + ",\n" + 
				" \"probability\": 100\n" + 
				"  }\n" + 
				" ]\n" + 
				"}";


		System.out.println(campaignBody);

		//Create a new campaign with our test case
		return connInfo.createCampaign(campaignBody);
	}
}