package com.cyara.apisamples;

import java.util.ArrayList;
import java.util.Date;

/**
 * Used to create a scheduled campaign run with test cases that have the 
 * same frequency and fall under the same runtimes. While this
 * class in and of itself doesn't create a campaign or insert
 * itself into one, we use it in {@link com.cyara.apisamples.SchedulePulseCampaigns}
 * to build and schedule campaigns with the most efficient combinations and schedules
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.0
 * @see com.cyara.apisamples.SchedulePulseCampaigns
 * @see com.cyara.apisamples.EligibleTestCase
 */
public class EligibleCampaignRun {
	private Date campaignRunStartDate;
	private Date campaignRunEndDate;
	private ArrayList<EligibleTestCase> eligibleTestCases;
	private int campaignFreq;
	private int requiredPortsForFrequencyRun;
	private int runDuration;
	private boolean runsMultipleTimes;
	private Date actualOneAndDoneEndDate;
	private int eligibleRunId;

	/**
	 * The constructor to create an EligibleCampaignRun object
	 * 
	 * @param campaignRunStartDate The mathematically calculated scheduled start time for this campaign run
	 * @param campaignRunEndDate The mathematically calculated scheduled end time for this campaign run
	 * @param eligibleTestCases The collection of test cases that will be run together as part of this run
	 * @param campaignFreq The Run Every for this campaign, is the same as each EligibleTestCase frequency
	 * @param requiredPortsForFrequencyRun The mathematically calculated smallest number of ports to achieve your campaignFreq
	 * @param runDuration The mathematically calculated number of minutes this campaign takes to run
	 * with the campaignFreq and calculated requiredPortsForFrequencyRun
	 * @param runsMultipleTimes If this campaign will execute more than once for this runtime
	 * @param actualOneAndDoneEndDate If the campaign doesn't run multiple times, it
	 * could be a regression monitoring suite, so we calculate when the one run actually ends
	 * @param eligibleRunId The unique ID we gave this run
	 */
	public EligibleCampaignRun(Date campaignRunStartDate, Date campaignRunEndDate, ArrayList<EligibleTestCase> eligibleTestCases, int campaignFreq, int requiredPortsForFrequencyRun, int runDuration, boolean runsMultipleTimes, Date actualOneAndDoneEndDate, int eligibleRunId) {
		this.campaignRunStartDate = campaignRunStartDate;
		this.campaignRunEndDate = campaignRunEndDate;
		this.eligibleTestCases = eligibleTestCases;
		this.campaignFreq = campaignFreq;
		this.requiredPortsForFrequencyRun = requiredPortsForFrequencyRun;
		this.runDuration = runDuration;
		this.runsMultipleTimes = runsMultipleTimes;
		this.actualOneAndDoneEndDate = actualOneAndDoneEndDate;
		this.eligibleRunId = eligibleRunId;
	}

	/**
	 * How often this campaign run should execute. Used when combining other
	 * eligible campaign runs together and when creating the campaign
	 * 
	 * @return The Run Every for this campaign run
	 * @since 1.0
	 */
	public int getCampaignFrequency() {
		return campaignFreq;
	}

	/**
	 * Retrieve the unique ID for this campaign run. Used to group together
	 * runs with the same configurations.
	 * 
	 * @return The unique ID we gave this run
	 * @since 1.0
	 */
	public int getId() {
		return eligibleRunId;
	}

	/**
	 * The common start time for this scheduled campaign run based on the
	 * eligible test cases this run has
	 * 
	 * @return The time the run's schedule starts for this campaign run
	 * @since 1.0
	 */
	public Date getStartDate() {
		return campaignRunStartDate;
	}

	/**
	 * The common end time for this scheduled campaign run based on the
	 * eligible test cases this run has
	 * 
	 * @return The time the run's schedule end for this campaign run
	 * @since 1.0
	 */
	public Date getEndDate() {
		return campaignRunEndDate;
	}

	/**
	 * The amount of ports this campaign run needs to run fast enough for the campaignFreq
	 * 
	 * @return The mathematically calculated smallest number of ports to achieve your campaignFreq
	 * @since 1.0
	 */
	public int getRequiredPortsForFrequencyRun() {
		return requiredPortsForFrequencyRun;
	}

	/**
	 * When building out this run, all the test cases that have the same
	 * frequency and have overlapping runtimes are grouped together. This
	 * is that group
	 * 
	 * @return The collection of test cases that will be run together as part of this run
	 * @since 1.0
	 */
	ArrayList<EligibleTestCase> getEligibleTestCases() {
		return eligibleTestCases;
	}

	/**
	 * When building out this run, all the test cases that have the same
	 * frequency and have overlapping runtimes are grouped together. This
	 * is that group
	 * 
	 * @return The mathematically calculated number of minutes this campaign takes to run
	 * with the campaignFreq and calculated requiredPortsForFrequencyRun
	 * @since 1.0
	 */
	public int getRunDuration() {
		return runDuration;
	}

	/**
	 * Campaigns can have a schedule of 5 minutes per day for a large set
	 * of test cases, especially when doing a regression monitoring suite.
	 * In Cyara, a campaign schedule just means when the campaign can launch.
	 * <p>
	 * We use this flag to determine if that's the case, and then calculate
	 * actualOneAndDoneEndDate
	 * 
	 * @return If this campaign will execute more than once for this runtime
	 * @since 1.0
	 */
	public boolean getRunsMultipleTimes() {
		return runsMultipleTimes;
	}

	/**
	 * This is only used if the campaign doesn't run multiple times, implying
	 * it's for a regression monitoring suite. Typically, a Pulse campaign is
	 * expected to end before the schedule run time ends.
	 * <p>
	 * But in Cyara, a campaign schedule just means when the campaign can launch. So the
	 * schedule run time end doesn't represent a hard stop. Thus, if this run is a
	 * regression monitoring suite and lasts longer than the schedule, we need to account
	 * for the ports this run will still be using and until when. This is that when.
	 * 
	 * @return The amount of time this campaign is actually going to run for
	 * @since 1.0
	 */
	public Date getActualOneAndDoneEndDate() {
		return actualOneAndDoneEndDate;
	}

	@Override
	public String toString() {
		Date endDateUsed = campaignRunEndDate;

		if (!runsMultipleTimes) {
			endDateUsed = actualOneAndDoneEndDate;
		}

		String testCaseIds = "";

		for (int x = 0; x < eligibleTestCases.size(); x++) {
			testCaseIds = testCaseIds + eligibleTestCases.get(x).getTestCaseId() + " ";
		}

		return "Frequency: " + campaignFreq + ". Ports: " + requiredPortsForFrequencyRun + ". Start rundatetime: " + campaignRunStartDate + ". End rundatetime: " + endDateUsed
				+ ". Test Case Ids: " + testCaseIds.trim() + ".";
	} 
}
