package com.cyara.apisamples;
import java.util.Date;

/**
 * Used for test cases that are currently supposed to run
 * for the currentDate we're on inside {@link com.cyara.apisamples.SchedulePulseCampaigns}.
 * If a test case is supposed to run based on the currentDate,
 * we create an object of this class which contains all the
 * relevant information about that test case and its runtime.
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.1
 * @see com.cyara.apisamples.SchedulePulseCampaigns
 * @see com.cyara.apisamples.EligibleCampaignRun
 */
public class EligibleTestCase implements Comparable<EligibleTestCase> {
	private int testCaseId;
	private int minuteDuration;
	private int frequency;
	private int scenarios;
	private String day;
	private String startTime;
	private String endTime;
	private Date startDate;
	private Date endDate;

	/**
	 * The constructor to create an EligibleTestCase object
	 * 
	 * @param testCaseId The portal's unique ID for the test case
	 * @param minuteDuration How long the test case takes to run, pulled from the description json
	 * @param frequency How often in minutes you want the test case to run, pulled from the description json
	 * @param scenarios How many data driven scenarios this test case has. If 0, is assigned 1 in {@link com.cyara.apisamples.SchedulePulseCampaigns} 
	 * @param day The day for this runtime, pulled from the description json
	 * @param startTime The time this test case needs to start running, pulled from the description json
	 * @param endTime The time this test case needs to stop running, pulled from the description json
	 * @param startDate The date representation of the day and startTime
	 * @param endDate The data representation of the day and endTime
	 */
	EligibleTestCase(int testCaseId, int minuteDuration, int frequency, int scenarios, String day, String startTime, String endTime, Date startDate, Date endDate) {
		this.testCaseId = testCaseId;
		this.minuteDuration = minuteDuration;
		this.frequency = frequency;
		this.scenarios = scenarios;
		this.day = day;
		this.startTime = startTime;
		this.endTime = endTime;
		this.startDate = startDate;
		this.endDate = endDate;
	}

	/**
	 * Gets the test case ID, which is used throughout {@link com.cyara.apisamples.SchedulePulseCampaigns}
	 * to group together test cases and, ultimately, create campaigns
	 * 
	 * @return The Portal's unique test case ID for this eligible test case
	 * @since 1.0
	 */
	public int getTestCaseId() {
		return testCaseId;
	}

	/**
	 * Gets the length in time this test case takes to run, which
	 * was specified in this test case's description json
	 * 
	 * @return The length in time this test case takes to run
	 * @since 1.0
	 */
	public int getMinuteDuration() {
		return minuteDuration;
	}

	/**
	 * Gets the length in time to wait between runs for this test
	 * case. Used in {@link com.cyara.apisamples.SchedulePulseCampaigns} to group
	 * test cases together that have the same frequency
	 * 
	 * @return How often in minutes you want this test case to run
	 * @since 1.0
	 */
	public int getTestCaseFrequency() {
		return frequency;
	}

	/**
	 * If the test case is data driven, every scenario is considered
	 * 1 call (so 8 scenarios would be 8 calls). If it's not data
	 * driven, then scenarios will just be 1, to represent how many
	 * calls this test case will ultimately make.
	 * 
	 * @return The number of calls this test case needs to make to run
	 * through all of its scenarios
	 * @since 1.0
	 */
	public int getScenarios() {
		return scenarios;
	}

	/**
	 * Based on the currentDate at the time this object was
	 * created, this test case had an eligible runtime for that
	 * date and time. This is the day in the description json
	 * that was used to determine that eligible runtime.
	 * 
	 * @return The day for this runtime, pulled from the description json
	 * @since 1.0
	 */
	public String getDay() {
		return day;
	}

	/**
	 * Based on the currentDate at the time this object was
	 * created, this test case had an eligible runtime for that
	 * date and time. This is the startTime in the description json
	 * that was used to determine that eligible runtime.
	 * 
	 * @return The time this test case needs to start running, pulled from the description json
	 * @since 1.0
	 */
	public String getStartTime() {
		return startTime;
	}

	/**
	 * Based on the currentDate at the time this object was
	 * created, this test case had an eligible runtime for that
	 * date and time. This is the endTime in the description json
	 * that was used to determine that eligible runtime.
	 * 
	 * @return The time this test case needs to stop running, pulled from the description json
	 * @since 1.0
	 */
	public String getEndTime() {
		return endTime;
	}

	/**
	 * The Date object representation of the day and start time
	 * for this test case runtime. Currently unused, but here for
	 * future convenience.
	 * 
	 * @return The date representation of the day and startTime
	 * @since 1.0
	 */
	public Date getTestCaseStartDate() {
		return startDate;
	}

	/**
	 * The Date object representation of the day and end time
	 * for this test case runtime. Currently unused, but here for
	 * future convenience.
	 * 
	 * @return The date representation of the day and endTime
	 * @since 1.0
	 */
	public Date getTestCaseEndDate() {
		return endDate;
	}

	/**
	 * Assigns the number of scenarios this test case has.
	 * For {@link com.cyara.apisamples.SchedulePulseCampaigns}, if a test
	 * case has no scenarios, it will set this to 1, because in
	 * that context, we use the scenarios to determine how many
	 * calls this test case will make.
	 * 
	 * @param scenarios Sets the number of calls this test case will make
	 * @since 1.0
	 */
	public void setScenarios(int scenarios) {
		this.scenarios = scenarios;
	}

	@Override
	public int compareTo(EligibleTestCase o) {
		return 0;
	}



}
