package com.cyara.apisamples;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Date;
import java.util.concurrent.ThreadLocalRandom;

import javax.json.JsonArray;
import javax.json.JsonObject;

import com.cyara.utilities.Utils;
import com.cyara.utilities.UtilsMath;
import com.cyara.utilities.UtilsTime;

/**
 * Grabs every Pulse call for a given timeframe on either
 * every attached account or the specified account. Generates
 * various metrics around connection times, response times, and more.
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.0
 */
public class IndustryStandard {
	public static void main(String[] argv) {
		//remainingSteps is structured as stepResultResult@@@stepResultDetail@@@responseTime
		//randomFakeAccNum~~~outputtedTestCaseId~~~runDateString~~~callResult~~~step0StepResult~~~step0StepResultDetail~~~step0ResponseTime~~~remainingSteps
		ApiFunctionalities connInfo = new ApiFunctionalities();
		collectData(connInfo, "2021-03-01T00:00:00.0000000Z", "2021-04-01T00:00:00.0000000Z", -1);
		processData(connInfo);
	}


	/**
	 * Pulls every Pulse call within the specified parameters,
	 * and anonymously parses out every basic data points for
	 * every call. Data points include response times,
	 * confidence scores, result, and more.
	 * 
	 * @param connInfo The controller to the API credentials and methods
	 * @param startString The date String of the earliest Pulse calls to grab
	 * @param endString The date String of the latest Pulse calls to grab
	 * @param accountNumber The value of which account to analyze. -1 to analyze all available
	 * @since 1.0
	 */
	private static void collectData(ApiFunctionalities connInfo, String startString, String endString, int accountNumber) {
		SimpleDateFormat apiFormat = new SimpleDateFormat("yyyy-MM-dd'T'HH:mm:ss.SSSSSSS'Z'");

		Date endDate = null;
		try {
			endDate = apiFormat.parse(endString);
		} catch (ParseException e) {
			e.printStackTrace();
			System.exit(1);
		}

		ArrayList<Integer> attachedAccountIds = new ArrayList<>();

		/*
		 * If this is -1, every attached account is grabbed.
		 * Otherwise, grabs just the specified account
		 */
		if (accountNumber == -1) {
			attachedAccountIds = getAttachedAccountIds(connInfo);
		}
		else {
			attachedAccountIds.add(accountNumber);
		}

		//Loop thru every account and collect data
		for (int accNumIndex = 0; accNumIndex < attachedAccountIds.size(); accNumIndex++) {
			int accNum = attachedAccountIds.get(accNumIndex);
			System.out.println("Processing account #" + accNum + " at index " + accNumIndex);

			//Change our connection engine to use the new account
			connInfo.setAccountNum(accNum);

			ArrayList<Integer> knownTestCaseIDs = new ArrayList<>();

			/*
			 * To keep any stored data anonymous when stored,
			 * a fake account number is given to the
			 * output file. This is that number.
			 */
			int randomFakeAccNum = ThreadLocalRandom.current().nextInt(10000, 2000000000 + 1);

			boolean morePages = true;
			String nextPage = "";

			//Get the initial Pulse feed for this account
			JsonObject apiObject = connInfo.getTestResultsFeed("pulse", "fromDate", startString);

			String pageTurner = "nextLink";

			//Main loop to keep grabbing more calls
			while (morePages) {
				if (apiObject.containsKey(pageTurner)) {
					String selfLink = apiObject.getString("selfLink");
					String nextLink = apiObject.getString(pageTurner);
					if (selfLink.equals(nextLink)) {
						morePages = false;
					}
					else {
						nextPage = apiObject.getString(pageTurner);
					}

					JsonArray resultsArr = apiObject.getJsonArray("results");
					for (int resultIndex = 0; resultIndex < resultsArr.size(); resultIndex++) {

						//Get specific call
						JsonObject previousResult = resultsArr.getJsonObject(resultIndex);

						String runDateString = previousResult.getString("runDate");

						/*
						 * Make sure the data is in the correct format.
						 * Certain results make the API drop the zeroes
						 */
						if (!runDateString.contains(".")) {
							runDateString = runDateString.replace("Z", ".0000000Z");
						}
						Date runDate = null;
						try {
							runDate = apiFormat.parse(runDateString);
						} catch (ParseException e) {
							e.printStackTrace();
							System.exit(1);
						}

						long timeBeforeTrackingEnd = runDate.getTime() - endDate.getTime();

						//Check that this call is within our date range
						if (timeBeforeTrackingEnd <= 0) {
							JsonObject testCase = previousResult.getJsonObject("testCase");
							int testCaseId = testCase.getInt("testCaseId");

							//If this is a new test case ID, store it for collection
							if (!knownTestCaseIDs.contains(testCaseId)) {
								knownTestCaseIDs.add(testCaseId);
							}

							int testCaseIdIndex = knownTestCaseIDs.indexOf(testCaseId);

							//Get the test result object of the call
							JsonObject callTestResult = previousResult.getJsonObject("testResult");

							//Get the result from the result object
							String callResult = callTestResult.getString("result");

							//Get the result detail from the result object
							//String callResultDetail = callTestResult.getString("detail");

							JsonObject testStepResultList = previousResult.getJsonObject("testStepResultList");

							String step0StepResult = "null";
							String step0StepResultDetail = "null";
							String step0ResponseTime = "null";

							//Collect connection metrics
							if (!testStepResultList.isNull("ringStepResult")) {
								JsonObject ringStepResult = testStepResultList.getJsonObject("ringStepResult");

								JsonObject ringStepResultStepResult = ringStepResult.getJsonObject("stepResult");

								step0StepResult = ringStepResultStepResult.getString("result");

								if (!ringStepResultStepResult.isNull("detail")) {
									step0StepResultDetail = ringStepResultStepResult.getString("detail");
								}

								step0ResponseTime = ringStepResult.getString("responseTime");
							}



							JsonArray callStepResultArr = testStepResultList.getJsonArray("callStepResult");

							StringBuilder remainingStepsBuilder = new StringBuilder();

							//Loop thru all other steps, and collect initiation and remaining step metrics
							for (int callStepIndex = 0; callStepIndex < callStepResultArr.size(); callStepIndex++) {
								JsonObject callStepResult = callStepResultArr.getJsonObject(callStepIndex);

								JsonObject step = callStepResult.getJsonObject("step");

								JsonObject expect = step.getJsonObject("expect");

								if (!expect.isNull("text") && expect.getString("text").length() > 0) {
									String responseTime = callStepResult.getString("responseTime");

									JsonObject stepResult = callStepResult.getJsonObject("stepResult");

									String stepResultResult = "null";
									if (!stepResult.isNull("result")) {
										stepResultResult = stepResult.getString("result");
									}

									String stepResultDetail = "null";
									if (!stepResult.isNull("detail")) {
										stepResultDetail = stepResult.getString("detail");
									}

									remainingStepsBuilder.append(stepResultResult + "@@@" + stepResultDetail + "@@@" + responseTime + "###");
								}
							}

							String remainingSteps = remainingStepsBuilder.toString();

							//Check if there were remaining steps, and thus trim the string
							if (remainingSteps.contains("#")) {
								remainingSteps = remainingSteps.substring(0, remainingSteps.length()-3);
							}

							Utils.logMessage(randomFakeAccNum + "~~~" + testCaseIdIndex + "~~~" + runDateString + "~~~" + callResult + "~~~" +
									step0StepResult + "~~~" + step0StepResultDetail + "~~~" + step0ResponseTime + "~~~" + remainingSteps, randomFakeAccNum + "_calls.log", true);

						}
						else {
							morePages = false;
							break;
						}

					}
				}
				else { //nextPage doesn't even exist, something is totally wrong
					morePages = false;
				}

				if (morePages) { //After processing that call, do we still have pages? Then make the next call here
					apiObject = connInfo.convertAPIOutputToJson(connInfo.makeApiGet(nextPage));
				}
			}
		}
	}


	/**
	 * Loop thru every account the user has access to, and
	 * store every single account ID.
	 * 
	 * @param connInfo The controller to the API credentials and methods
	 * @return An ArrayList with all the account IDs the user is attached to
	 * @since 1.0
	 */
	private static ArrayList<Integer> getAttachedAccountIds(ApiFunctionalities connInfo) {
		ArrayList<Integer> attachedAccountIds = new ArrayList<>();

		boolean morePages = true;
		String nextPage = "";

		JsonObject attachedAccountsObj = connInfo.listAccounts();

		String pageTurner = "nextPage";

		//Loop through all the pages of listed accounts, storing all account IDs
		while (morePages) {
			if (attachedAccountsObj.containsKey(pageTurner)) {
				if (attachedAccountsObj.isNull(pageTurner)) {
					morePages = false;
				}
				else {
					nextPage = attachedAccountsObj.getString(pageTurner);
				}

				JsonArray accountArr = attachedAccountsObj.getJsonArray("results");
				for (int accountIndex = 0; accountIndex < accountArr.size(); accountIndex++) {
					JsonObject account = accountArr.getJsonObject(accountIndex);
					int accountId = account.getInt("accountId");
					attachedAccountIds.add(accountId);
				}

			}
			else { //nextPage doesn't even exist, something is totally wrong
				morePages = false;
			}

			if (morePages) { //After processing that call, do we still have pages? Then make the next call here
				attachedAccountsObj = connInfo.convertAPIOutputToJson(connInfo.makeApiGet(nextPage));
			}
		}

		return attachedAccountIds;
	}


	/**
	 * Loop thru every Pulse call for every account collected for
	 * the given timeframe. Categorize and store the various
	 * basic metrics, then process and output them.
	 * 
	 * @param connInfo The controller to the API credentials and methods
	 * @since 1.0
	 */
	private static void processData(ApiFunctionalities connInfo) {
		try {

			File myFolder = new File(".");

			File[] fileArr = myFolder.listFiles();

			int totalCallCounter = 0;
			ArrayList<Integer> totalTCsPerAccount = new ArrayList<>();
			ArrayList<Integer> allConnTimes = new ArrayList<>();
			ArrayList<Integer> allInitTimes = new ArrayList<>();
			ArrayList<Integer> remainingResponseTimes = new ArrayList<>();
			ArrayList<Double> allConfScores = new ArrayList<>();
			ArrayList<Double> normalizedSuccessPcts = new ArrayList<>();

			/*
			 * Start looping thru every file of every account stored.
			 * Categorize and massage the key data points for analysis.
			 */
			for (int fileIndex = 0; fileIndex < fileArr.length; fileIndex++) {
				if (fileArr[fileIndex].isFile()) {
					String fileName = fileArr[fileIndex].getName();
					if (fileName.endsWith("_calls.log")) {
						int accountCallCounter = 0;
						int accountSuccessCallCounter = 0;
						int largestAccountTestCaseId = -1;

						String fileLine = "";

						//Loop thru the file of every call collected for this account
						try (BufferedReader br = new BufferedReader(new FileReader(fileName))) {
							while ((fileLine = br.readLine()) != null) {
								String[] lineArr = fileLine.split("~~~", -1);
								try {
									int testCaseIdIndex = Integer.parseInt(lineArr[1]);
									if (testCaseIdIndex > largestAccountTestCaseId) {
										largestAccountTestCaseId = testCaseIdIndex;
									}

									String step0StepResultDetail = lineArr[5];
									String step0StepResult = lineArr[4];

									if (!step0StepResultDetail.contains("No answer") && !step0StepResultDetail.contains("invalid format") &&
											!step0StepResult.equals("InternalError") && !step0StepResult.equals("Aborted")) {
										allConnTimes.add(UtilsTime.getMillis(lineArr[6]));
									}

									if (step0StepResult.equals("Success")) {
										accountSuccessCallCounter++;
									}

									/*
									 * Check if this call has more than just step 0.
									 * If it does, start processing all those other steps.
									 */
									if (lineArr.length == 8) {
										try {
											String[] remainingStepsArr = lineArr[7].split("###");

											if (remainingStepsArr[0].length() > 1) {

												//Parse out and store the Step 1 response time
												String[] initStep = remainingStepsArr[0].split("@@@");
												int initMillis = UtilsTime.getMillis(initStep[2]);
												if (initMillis > 1) {
													allInitTimes.add(initMillis);
												}

												//Loop thru every Step 1 and beyond and store the confidence scores
												for (int remainingSteps = 0; remainingSteps < remainingStepsArr.length; remainingSteps++) {
													String[] remainingStepArr = remainingStepsArr[remainingSteps].split("@@@");

													String stepResultDetail = remainingStepArr[1];
													if (stepResultDetail.contains("onfidence")) {
														if (stepResultDetail.contains("nhanced confidence ")) {
															stepResultDetail = stepResultDetail.substring(stepResultDetail.indexOf("onfidence ") + 10, stepResultDetail.length());
														}
														else {
															stepResultDetail = stepResultDetail.substring(stepResultDetail.indexOf("onfidence: ") + 11, stepResultDetail.length());
														}
														stepResultDetail = stepResultDetail.substring(0, stepResultDetail.indexOf("%"));
														try {
															allConfScores.add(Double.parseDouble(stepResultDetail.trim()));
														}
														catch (Exception ex) {
															System.out.println(fileLine);
															ex.printStackTrace();
															System.exit(1);
														}

													}

												}
											}

											//We need to check if there's more than 1 step to process the remaining response times...
											if (remainingStepsArr.length > 1) {

												//...because we're starting at step 2 (index 1 of array)
												for (int remainingSteps = 1; remainingSteps < remainingStepsArr.length; remainingSteps++) {
													String[] remainingStepArr = remainingStepsArr[remainingSteps].split("@@@");

													int remainingResponseMillis = UtilsTime.getMillis(remainingStepArr[2]);

													if (remainingResponseMillis > 1) {
														remainingResponseTimes.add(remainingResponseMillis);
													}
												}
											}

										}
										catch (Exception ex) {
											//System.out.println(fileArr[fileIndex].getName());
											//System.out.println(fileLine);
											//ex.printStackTrace();
											//System.exit(1);
										}
									}

									totalCallCounter++;
									accountCallCounter++;
								}
								catch (Exception ex) {
									//System.out.println(fileArr[fileIndex].getName());
									//System.out.println(fileLine);
									//ex.printStackTrace();
									//System.exit(1);
								}

							}
						}

						normalizedSuccessPcts.add(UtilsMath.generateCleanPercentages(accountSuccessCallCounter + 1, accountCallCounter + 2));
						totalTCsPerAccount.add(largestAccountTestCaseId + 1);
					}
				}
			}

			Collections.sort(totalTCsPerAccount);
			Collections.sort(allConnTimes);
			Collections.sort(allInitTimes);
			Collections.sort(remainingResponseTimes);
			Collections.sort(allConfScores);
			Collections.sort(normalizedSuccessPcts);

			System.out.println("Total number of Pulse calls processed for March 2021: " + totalCallCounter);

			outputData(totalTCsPerAccount, allConnTimes, allInitTimes, remainingResponseTimes, allConfScores, normalizedSuccessPcts);

		}
		catch (Exception ex) {
			ex.printStackTrace();
		}
	}


	/**
	 * Initiate the calculations and print out the results for all
	 * the Pulse metrics collected.
	 * 
	 * @param totalTCsPerAccount Every number of test cases by account
	 * @param allConnTimes Every Step 0 response time collected
	 * @param allInitTimes Every Step 1 response time collected
	 * @param remainingResponseTimes Every Step 2 and beyond response time collected
	 * @param allConfScores Every confidence score result collected
	 * @param normalizedSuccessPcts Every account success percentage
	 * @since 1.0
	 */
	private static void outputData(ArrayList<Integer> totalTCsPerAccount, ArrayList<Integer> allConnTimes,
			ArrayList<Integer> allInitTimes, ArrayList<Integer> remainingResponseTimes, ArrayList<Double> allConfScores,
			ArrayList<Double> normalizedSuccessPcts) {

		System.out.println();

		System.out.println("Mean number of Pulse test cases per account: " + UtilsMath.findMean(totalTCsPerAccount));
		System.out.println("Median number of Pulse test cases per account: " + UtilsMath.findMedian(totalTCsPerAccount));
		System.out.println("Mode number of Pulse test cases per account: " + UtilsMath.findMode(totalTCsPerAccount));
		System.out.println("95 percentile number of Pulse test cases per account: " + UtilsMath.findPercentile(totalTCsPerAccount, 95.0));
		System.out.println("Largest number of Pulse test cases uses for a single account: " + totalTCsPerAccount.get(totalTCsPerAccount.size()-1));

		System.out.println();

		System.out.println("Mean normalized success percentage of Pulse calls per account: " + UtilsMath.findMeanDouble(normalizedSuccessPcts) + "%");
		System.out.println("Median normalized success percentage of Pulse calls per account: " + UtilsMath.findMedianDouble(normalizedSuccessPcts) + "%");
		System.out.println("Mode normalized success percentage of Pulse calls per account: " + UtilsMath.findModeDouble(normalizedSuccessPcts) + "%");
		System.out.println("95 percentile normalized success percentage of Pulse calls per account: " + UtilsMath.findPercentileDouble(normalizedSuccessPcts, 95.0) + "%");
		System.out.println("Highest normalized success percentage of Pulse calls for a single account: " + normalizedSuccessPcts.get(normalizedSuccessPcts.size()-1) + "%");

		System.out.println();

		System.out.println("Mean length of connection time in ms: " + UtilsMath.findMean(allConnTimes));
		System.out.println("Median length of connection time in ms: " + UtilsMath.findMedian(allConnTimes));
		System.out.println("Mode length of connection time in ms: " + UtilsMath.findMode(allConnTimes));
		System.out.println("95 percentile length of connection time in ms: " + UtilsMath.findPercentile(allConnTimes, 95.0));
		System.out.println("Longest connection time in ms: " + allConnTimes.get(allConnTimes.size()-1));

		System.out.println();

		System.out.println("Mean length of initiation time in ms: " + UtilsMath.findMean(allInitTimes));
		System.out.println("Median length of initiation time in ms: " + UtilsMath.findMedian(allInitTimes));
		System.out.println("Mode length of initiation time in ms: " + UtilsMath.findMode(allInitTimes));
		System.out.println("95 percentile length of initiation time in ms: " + UtilsMath.findPercentile(allInitTimes, 95.0));
		System.out.println("Longest initiation time in ms: " + allInitTimes.get(allInitTimes.size()-1));

		System.out.println();

		System.out.println("Mean length of remaining response times per step in ms: " + UtilsMath.findMean(remainingResponseTimes));
		System.out.println("Median length of remaining response times per step in ms: " + UtilsMath.findMedian(remainingResponseTimes));
		System.out.println("Mode length of remaining response times per step in ms: " + UtilsMath.findMode(remainingResponseTimes));
		System.out.println("95 percentile length of remaining response times per step in ms: " + UtilsMath.findPercentile(remainingResponseTimes, 95.0));
		System.out.println("Longest remaining response times per step in ms: " + remainingResponseTimes.get(remainingResponseTimes.size()-1));

		System.out.println();

		System.out.println("Mean confidence score of every prompt: " + UtilsMath.findMeanDouble(allConfScores) + "%");
		System.out.println("Median confidence score of every prompt: " + UtilsMath.findMedianDouble(allConfScores) + "%");
		System.out.println("Mode confidence score of every prompt: " + UtilsMath.findModeDouble(allConfScores) + "%");
		System.out.println("95 percentile confidence score of every prompt: " + UtilsMath.findPercentileDouble(allConfScores, 95.0) + "%");
	}

}
