package com.cyara.apisamples;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.nio.file.Files;
import java.nio.file.Paths;

import javax.json.JsonArray;
import javax.json.JsonObject;

import com.cyara.utilities.*;

/**
 * Grabs all the Pulse calls for a given timeframe and checks for
 * any that failed on Step 0 or Step 1, a connection failure. Then
 * checks transcriptions and audio recordings to find the closest
 * matching failure, broken down by each. These failures and matches
 * are then process recursively into a hierarchal network graph,
 * showing you the most common failure types. 
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.0
 */
public class MatchedConnectionFailures {
	private static final String OUTPUTFILENAME = "FailedResults.txt";
	private static final String MATCHESFILENAME = "FailedResultsWithMatches.txt";

	public static void main(String[] argv) {

		findStep0Or1Failures();

		addClosestMatches();

		//Create Transcriptions Network Graph HTML File
		Utils.logMessage(GenerateNetworkGraph.generate(MATCHESFILENAME, "closestTranscriptionTestResultId", 30, 20), "TranscriptionsNetworkGraph.html", false);

		//Create Audio Recordings Network Graph HTML File
		Utils.logMessage(GenerateNetworkGraph.generate(MATCHESFILENAME, "closestAudioFile", 15, 10), "AudioNetworkGraph.html", false);

	}


	/**
	 * Grabs all the Pulse calls from a selected date until now.
	 * Each Pulse call is processed, checking for a Step 0 or
	 * Step 1 Failure. If it is, the call is stored. If the
	 * failure is on Step 1, the audio for Step 1 is downloaded.
	 *  
	 * @since 1.0
	 */
	private static void findStep0Or1Failures() {
		ApiFunctionalities connInfo = new ApiFunctionalities();

		Utils.logMessage("runDate~testCaseId~testCaseName~destination~testResultId~failedStepNumber~stepResultDetail~transcription~hasRecording~closestTranscriptionTestResultId~closestDiff~closestMatchPct~closestAudioFile~highestMos", OUTPUTFILENAME, true);

		boolean morePages = true;
		String nextPage = "";

		JsonObject apiObject = connInfo.getTestResultsFeed("pulse", "fromDate", "2021-04-24T00:00:00Z");

		//Main loop to keep grabbing more calls
		while (morePages) {
			if (apiObject.containsKey("nextLink")) {
				String selfLink = apiObject.getString("selfLink");
				String nextLink = apiObject.getString("nextLink");
				if (selfLink.equals(nextLink)) {
					morePages = false;
				}
				else {
					nextPage = apiObject.getString("nextLink");
				}

				JsonArray resultsArr = apiObject.getJsonArray("results");
				for (int resultIndex = 0; resultIndex < resultsArr.size(); resultIndex++) {
					//Get specific call
					JsonObject call = resultsArr.getJsonObject(resultIndex);

					//Get the test result object of the call
					JsonObject callTestResult = call.getJsonObject("testResult");

					//Get the result from the result object
					String callResult = callTestResult.getString("result");

					if (callResult.equals("Failed")) {
						String logCallMsg = "";
						//In here, further process previousResult by step
						String runDate = call.getString("runDate");
						int testResultId = call.getInt("testResultId");
						String destination = call.getString("calledNumber");

						JsonObject testCase = call.getJsonObject("testCase");

						String testCaseName = testCase.getString("name");
						int testCaseId = testCase.getInt("testCaseId");

						logCallMsg = runDate + "~" + testCaseId + "~" + testCaseName + "~" + destination + "~" + testResultId + "~";

						JsonObject testStepResultList = call.getJsonObject("testStepResultList");

						JsonObject ringStepResult = testStepResultList.getJsonObject("ringStepResult");

						JsonObject ringStepResultBreakdown = ringStepResult.getJsonObject("stepResult");
						String ringStepResultResult = ringStepResultBreakdown.getString("result");
						String ringStepResultDetail = ringStepResultBreakdown.getString("detail");
						if (ringStepResultResult.equals("Failed")) {
							logCallMsg = logCallMsg + "0~" + ringStepResultDetail + "~" + "~~~" + "~" + "~";
							Utils.logMessage(logCallMsg, OUTPUTFILENAME, true);
						}
						else {
							JsonArray callStepResult = testStepResultList.getJsonArray("callStepResult");

							if (!callStepResult.isEmpty()) {
								JsonObject stepDetails = callStepResult.getJsonObject(0);

								JsonObject stepResultBreakdown = stepDetails.getJsonObject("stepResult");
								String stepResultResult = stepResultBreakdown.getString("result");
								String stepResultDetail = stepResultBreakdown.getString("detail");

								if (stepResultResult.equals("Failed")) {
									JsonObject step1TranscriptionJson = connInfo.generateStepTranscription(testResultId, 1);
									String step1Transcription = step1TranscriptionJson.getString("transcription");
									System.out.println(step1Transcription);
									//Utils.sleepMode(1200);

									JsonObject stepsRecordingsAvailable = connInfo.getStepRecordingAvailabilities(testResultId);
									JsonArray recordedSteps = stepsRecordingsAvailable.getJsonArray("results");
									JsonObject step1RecordingAvailability = recordedSteps.getJsonObject(1);
									boolean hasStep1Recording = step1RecordingAvailability.getBoolean("stepRecording");

									if (hasStep1Recording) {
										connInfo.getRecording(testResultId, 1);
									}

									logCallMsg = logCallMsg + "1~" + stepResultDetail + "~" + step1Transcription + "~" + hasStep1Recording + "~";

									Utils.logMessage(logCallMsg, OUTPUTFILENAME, true);
								}
							}
						}
					}
				}
			}
			else { //nextPage doesn't even exist, something is totally wrong
				morePages = false;
			}

			if (morePages) { //After processing that call, do we still have pages? Then make the next call here
				apiObject = connInfo.convertAPIOutputToJson(connInfo.makeApiGet(nextPage));
			}
		}
	}


	/**
	 * Processes the file of Step 0 and 1 Failures, finding similar
	 * matching failures for each one. All of this is then exported
	 * to a new file, which GenerateNetworkGraph will then use.
	 * 
	 * @see findClosestAudio
	 * @see findClosestTranscription
	 * @since 1.0
	 */
	private static void addClosestMatches() {
		String fileLine = "";
		int testResultIdIndex = -1;
		int transcriptionTestResultIdIndex = -1;
		int hasRecordingIndex = -1;

		//Loop thru the list of failures for processing
		try (BufferedReader br = new BufferedReader(new FileReader(OUTPUTFILENAME))) {
			while ((fileLine = br.readLine()) != null) {
				StringBuilder fileLineSB = new StringBuilder(fileLine);

				//Split the row into its separated columns
				String[] failureInfo = fileLine.split("~");

				//testCaseId means we're on the first row, so grab the header indices
				if (fileLine.contains("testCaseId")) {
					testResultIdIndex = Utils.getHeaderIndex(failureInfo, "testResultId");
					transcriptionTestResultIdIndex = Utils.getHeaderIndex(failureInfo, "transcription");
					hasRecordingIndex = Utils.getHeaderIndex(failureInfo, "hasRecording");
				}
				else {
					int testResultId = Integer.parseInt(failureInfo[testResultIdIndex]);
					String transcription = failureInfo[transcriptionTestResultIdIndex];
					boolean hasRecording = Boolean.parseBoolean(failureInfo[hasRecordingIndex]);

					//Find and append the closest matching transcription
					fileLineSB.append(findClosestTranscription(transcription, testResultId) + "~");

					//Find and append the closest matching recording
					if (hasRecording) {
						fileLineSB.append(findClosestAudio(testResultId + "Step1.wav"));
					}
					else {
						fileLineSB.append("~");
					}
				}

				/*
				 * Append the failure row, which now has
				 * any matching data, to the new file.
				 */
				Utils.logMessage(fileLineSB.toString(), MATCHESFILENAME, true);
			}
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}

	}


	/**
	 * Processes a specific failed call's audio, running PESQ
	 * against every other failure audio to find the closest match.
	 * 
	 * @param step1AudioFilePath This failure's audio file
	 * @return The closest matching failure audio, if any, and its MOS
	 * @since 1.0
	 */
	private static String findClosestAudio(String step1AudioFilePath) {
		String closestAudioFile = "";
		String pesqLine = "";
		double highestMos = -1;

		// Path of folder where files are located 
		String folderPath = ".";

		// creating new folder 
		File myfolder = new File(folderPath); 

		File[] fileArr = myfolder.listFiles(); 
		for (int i = 0; i < fileArr.length; i++) { 
			if (fileArr[i].isFile()) {
				String oldFileName = fileArr[i].getName();

				/*
				 * This other failure has a recording, and it isn't
				 * the same as the one we're actively processing
				 */
				if (oldFileName.endsWith(".wav") && !oldFileName.equals(step1AudioFilePath)) {
					try {

						/*
						 * Run the PESQ algorithm to comparing the
						 * processed failure to this other failure.
						 * 
						 * If you are running a windows command,
						 * you need to put cmd /c in front of your command.
						 */
						Process pr = Runtime.getRuntime().exec("PESQ +8000 " + step1AudioFilePath + " " + oldFileName);
						pr.waitFor();
						BufferedReader input = new BufferedReader(new InputStreamReader(pr.getInputStream()));

						/*
						 * Get the output from the PESQ algorithm,
						 * and find the line containing the MOS.
						 */
						while ((pesqLine = input.readLine()) != null) {
							if (pesqLine.contains("P.862 Prediction")) {
								break;
							}
						}
						input.close();

						/*
						 * Parse out the MOS from that PESQ line,
						 * and check if it's the new closest match.
						 */
						try {
							double mos = Double.parseDouble((pesqLine.substring(pesqLine.indexOf("	")+1, pesqLine.length())));

							System.out.println("The MOS for the two files was: " + mos);

							if (mos > highestMos) {
								highestMos = mos;
								closestAudioFile = oldFileName;
							}
						}
						catch (NullPointerException ex) {

						}
					} catch (IOException e) {
						e.printStackTrace();
					} catch (InterruptedException e) {
						e.printStackTrace();
						Thread.currentThread().interrupt();
					}
				}
			}
		}

		/*
		 * The closest audio match has a MOS higher than 3,
		 * meaning it's worthy of being a match and not
		 * just noise. So we return it and the score.
		 */
		if (highestMos > 3) {
			try {
				Files.delete(Paths.get(step1AudioFilePath));
			} catch (IOException e) {
				e.printStackTrace();
				System.exit(1);
			}
			return closestAudioFile + "~" + highestMos;
		}
		else {
			return "~";
		}

	}


	/**
	 * Processes a specific failed call's transcription, running
	 * Levenshtein against every other failure transcription to find
	 * the closest match.
	 * 
	 * @param step1Transcription This failure's transcription
	 * @param testResultId The id of the failure we're processing
	 * @return The closest matching testResultId, and how close
	 * @since 1.0
	 */
	private static String findClosestTranscription(String step1Transcription, int testResultId) {
		int step1Len = step1Transcription.length();
		int closestTranscriptionTestResultId = -1;

		int closestDiff = 999999;
		double closestMatchPct = -1;

		String fileLine = "";

		int prevTranscriptionTestResultIdIndex = -1;
		int prevTestResultIdIndex = -1;

		//Verify this failure even has a transcription
		if (step1Len > 1) {

			//Loop thru the failures file and start finding the closest match
			try (BufferedReader br = new BufferedReader(new FileReader(OUTPUTFILENAME))) {
				while ((fileLine = br.readLine()) != null) {
					//Split the row into its separated columns
					String[] prevFailureInfo = fileLine.split("~");

					//testCaseId means we're on the first row, so grab the header indices
					if (fileLine.contains("testCaseId")) {
						prevTestResultIdIndex = Utils.getHeaderIndex(prevFailureInfo, "testResultId");
						prevTranscriptionTestResultIdIndex = Utils.getHeaderIndex(prevFailureInfo, "transcription");
					}
					else {
						//Get the transcription for this other failure
						String prevTranscription = prevFailureInfo[prevTranscriptionTestResultIdIndex];

						//Get the testResultId for this other failure
						int prevTestResultId = Integer.parseInt(prevFailureInfo[prevTestResultIdIndex]);

						if (testResultId != prevTestResultId) {

							/*
							 * Get the difference in characters between the
							 * original failure's transcription and this
							 * other failure's transcription...
							 */
							int diff = UtilsMath.distance(step1Transcription, prevTranscription);

							//...and get the percentage that difference represents
							double diffPct = UtilsMath.generateCleanPercentages(diff, step1Len);

							/*
							 * Check if this other failure is a closer
							 * match than our previously closest
							 * failure and overwrite
							 */
							if (diff < closestDiff) {
								closestDiff = diff;
								closestMatchPct = diffPct;
								closestTranscriptionTestResultId = prevTestResultId;
							}
						}
					}
				}
			} catch (IOException e) {
				e.printStackTrace();
			}
		}

		return closestTranscriptionTestResultId + "~" + closestDiff + "~" + closestMatchPct;
	}


}
