package com.cyara.apisamples;

import javax.json.JsonObject;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

/**
 * Takes in a test case ID and generates a cleaner, simpler table
 * of a test case with all the expect to hears and replies. The
 * table is HTML with inline CSS and is printed out to the console.
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.1
 */
public class PrettyPrintTestCase {
	public static void main(String[] argv) {
		ApiFunctionalities connInfo = new ApiFunctionalities();
		int testCaseId = -1; //Change to your test case ID

		//Call the central table generation method and print it out
		String prettyTestCaseHTML = generatePrettyPrint(connInfo, testCaseId);
		System.out.println(prettyTestCaseHTML);

	}


	/**
	 *  The central method that launches the various HTML
	 *  builders and compiles and constructs the final table.
	 *  
	 *  @param connInfo The controller to the API credentials and methods
	 *  @param testCaseId The id we're generating the table for
	 *  @return The final HTML String of the table
	 *  @since 1.0
	 */
	private static String generatePrettyPrint(ApiFunctionalities connInfo, int testCaseId) {

		//Start building out the test case basic info table
		String[] testCaseTableHeaders = {"Test Case ID", "Name", "Used Number", "Desc.", "Notes"};

		//Calculate the column widths for the basic info table
		int[] testCaseTableHeaderWidths = {-1, -1, 90, -1, -1};
		int[] testCaseTableColumnWidths = calcColWidths(testCaseTableHeaderWidths, testCaseTableHeaders);

		String testCaseTableHeading = generateTableHead(testCaseTableColumnWidths, testCaseTableHeaders);

		String testCaseTableRows = "    <tbody>\n";

		StringBuilder masterTestCaseBreakdownTableBuilder = new StringBuilder();

		//Get the test case basics to start processing
		JsonObject testCaseBasics = connInfo.getTestCaseBasics(testCaseId);

		String testCaseName = testCaseBasics.getString("name");

		String calledNumber = testCaseBasics.getString("calledNumber");

		String testCaseDescription = "";
		if (!testCaseBasics.isNull("description")) {
			testCaseDescription = testCaseBasics.getString("description");
		}

		String notes = "";
		if (!testCaseBasics.isNull("notes")) {
			notes = testCaseBasics.getString("notes");
		}

		//Build the one and only row for the basic info table, containing the basic info we just grabbed
		String[] testCaseTableRow = {String.valueOf(testCaseId), testCaseName, calledNumber, testCaseDescription, notes};
		testCaseTableRows = testCaseTableRows + generateTableRow(testCaseTableColumnWidths, testCaseTableRow, 1);



		//Start building out the test case step breakdown table
		String[] testCaseBreakdownHeaders = {"Step No.", "Desc.", "Expect to Hear", "Reply With"};

		//Calculate the column widths for the breakdown table
		int[] testCaseBreakdownHeaderWidths = {32, -1, -1, -1};
		int[] testCaseBreakdownColumnWidths = calcColWidths(testCaseBreakdownHeaderWidths, testCaseBreakdownHeaders);

		String testCaseBreakdownHeading = generateTableHead(testCaseBreakdownColumnWidths, testCaseBreakdownHeaders);

		String testCaseBreakdownRows = "    <tbody>\n";

		String[] step0Row = {"0", "Connection / Answering", "", ""};
		testCaseBreakdownRows = testCaseBreakdownRows + generateTableRow(testCaseBreakdownColumnWidths, step0Row, 1);

		//Export this test case as XML to start processing all the steps
		String rawXML = connInfo.exportTestCase(testCaseId, "FlattenBlocks");

		Document sampleDoc = connInfo.convertStringToDocument(rawXML);

		sampleDoc.getDocumentElement().normalize();
		NodeList nodeList = sampleDoc.getElementsByTagName("Step");

		//Loop thru all the steps in the test case, generating the table row with their ETH and Reply
		for (int temp = 0; temp < nodeList.getLength(); temp++) {
			org.w3c.dom.Node node = nodeList.item(temp);

			if (node.getNodeType() == org.w3c.dom.Node.ELEMENT_NODE) {
				Element element = (Element) node;
				String stepDescription = element.getElementsByTagName("Description").item(0).getTextContent();
				String expectedText = element.getElementsByTagName("ExpectedText").item(0).getTextContent();
				String replyText = element.getElementsByTagName("ReplyText").item(0).getTextContent();


				String[] testCaseBreakdownRow = {String.valueOf(temp+1), stepDescription, expectedText, replyText};
				testCaseBreakdownRows = testCaseBreakdownRows + generateTableRow(testCaseBreakdownColumnWidths, testCaseBreakdownRow, (temp+2));
			}
		}


		//Finalize the test case basic info table by building the footer and assembling all parts
		String testCaseTableFoot = generateTableFoot("Test Case to Test", testCaseTableHeaders.length);

		String testCaseTableFinal = testCaseTableHeading + testCaseTableRows + testCaseTableFoot;


		//Finalize the test case step breakdown table by bulding the footer and assembling all parts
		String testCaseBreakdownFoot = generateTableFoot("Test Case ID " + testCaseId, testCaseBreakdownHeaders.length);
		masterTestCaseBreakdownTableBuilder.append(testCaseBreakdownHeading + testCaseBreakdownRows + testCaseBreakdownFoot);

		return "<!DOCTYPE html>\n" + 
		"<html>\n" + 
		"<body>\n" + testCaseTableFinal + masterTestCaseBreakdownTableBuilder.toString() + "</body>\n" + 
		"</html>";
	}


	/**
	 *  Generates the HTML for the header columns, including
	 *  their color. Uses the passed in widths, and content.
	 *  
	 *  @param colWidths The pixel width for each header column
	 *  @param headers The id we're generating the table for
	 *  @return The HTML for the table's header
	 *  @since 1.0
	 */
	public static String generateTableHead(int[] colWidths, String[] headers) {

		String headColumns = "";

		for (int x = 0; x < colWidths.length; x++) {
			headColumns = headColumns + "            <th style=\"background: #289FCA;color: white;font-weight: bold;padding: 10px;border: 1px solid #ccc;text-align: left;font-size: 14px;max-width: " + colWidths[x] + "px;word-wrap: break-word;font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif;\">" + headers[x] + "</th>\n";
		}

		String tableHeader = "<table style=\"border-collapse: collapse;margin: 10px auto;font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif;\">\n" + 
				"    <thead>\n" + 
				"        <tr>\n" + 
				headColumns +
				"        </tr>\n" + 
				"    </thead>\n";

		return tableHeader;
	}


	/**
	 *  Generates the HTML for the specified step row, including
	 *  their color. Uses the passed in widths, and content.
	 *  
	 *  @param colWidths The pixel width for each column
	 *  @param rowCells The string content for each cell of the row
	 *  @param rowNumber The index of which row we're
	 *  processing. Used for background shading
	 *  @return The HTML for the row
	 *  @since 1.0
	 */
	public static String generateTableRow(int[] colWidths, String[] rowCells, int rowNumber) {
		String tableRow = "        <tr>\n";
		String bgColor = "";

		if (rowNumber % 2 == 0) {
			bgColor = "background: #F4F4F4;";
		}

		for (int x = 0; x < rowCells.length; x++) {
			tableRow = tableRow + "            <td style=\"padding: 10px;border: 1px solid #ccc;text-align: left;font-size: 14px;max-width: " + colWidths[x] + "px;word-wrap: break-word;font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif;" + bgColor + "\">" + rowCells[x] + "</td>\n";
		}

		tableRow = tableRow + "        </tr>\n";

		return tableRow;
	}


	/**
	 *  Generates the HTML for the table's footer where we
	 *  display the table's label.
	 *  
	 *  @param tableLabel The unique label shown underneath the table for
	 *  what this table contains
	 *  @param columnCount The number of columns this table has. Used
	 *  to give the correct width
	 *  @return The HTML for the table's footer
	 *  @since 1.0
	 */
	public static String generateTableFoot(String tableLabel, int columnCount) {
		return 				"    <tfoot>\n" + 
				"         <tr>\n" + 
				"           <td colspan=\"" + columnCount + "\" style=\"padding: 5px;border: 0px solid #ccc;text-align: center;font-size: 14px; background: #FFFFFF;max-width: 100px;word-wrap: break-word;color: #165DAA;font-family: 'Helvetica Neue',Helvetica,Arial,sans-serif;\">" + tableLabel + "</td>\n" + 
				"         </tr>\n" + 
				"    </tfoot>\n" + 
				"</table><br></br><br></br>\n";
	}


	/**
	 *  Calculates the correct column widths for the supplied values.
	 *  Uses the number of headers being passed, and any desired column
	 *  widths to calculate the equal number of column widths to be used
	 *  for the remaining columns.
	 *  
	 *  @param predeterminedHeaderWidths The passed header widths. Any specified,
	 *  predetermined header widths have a value other than -1. The -1 widths will get
	 *  assigned the equally divided remaining width.
	 *  @param headers The headers this table contains, only used for their count
	 *  @return The array of all the column widths
	 *  @since 1.0
	 */
	public static int[] calcColWidths(int[] predeterminedHeaderWidths, String[] headers) {
		int totalHeaders = headers.length;

		//The maximum pixel width this table can be, likely won't need to change
		int maxWidthTotal = 400;
		int dynamicCount = 0;

		int[] headerWidths = new int[totalHeaders];

		//Loop thru the passed widths, and subtract from the max width how many remaining pixels we have
		for (int x = 0; x < predeterminedHeaderWidths.length; x++) {
			int predeterminedWidth = predeterminedHeaderWidths[x];

			if (predeterminedWidth != -1) {
				maxWidthTotal -= predeterminedWidth;
			}
			else {
				dynamicCount++;
			}
		}

		//Evenly divide the remaining width by the number of columns with unspecified widths
		int dynamicWidth = 1;

		if (dynamicCount > 0) {
			dynamicWidth = maxWidthTotal / dynamicCount;
		}

		//Loop thru all the columns and assign the equal remaining width to unspecified columns
		for (int x = 0; x < predeterminedHeaderWidths.length; x++) {
			int predeterminedWidth = predeterminedHeaderWidths[x];

			if (predeterminedWidth != -1) {
				headerWidths[x] = predeterminedWidth;
			}
			else {
				headerWidths[x] = dynamicWidth;
			}
		}

		return headerWidths;
	}

}
