package com.cyara.apisamples;

import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.time.Instant;
import java.time.format.DateTimeFormatter;
import java.time.temporal.ChronoUnit;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;

import javax.json.JsonArray;
import javax.json.JsonObject;

import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

/**
 * A program to completely automate the creation and scheduling
 * of your Pulse campaigns. Will calculate the most efficient
 * schedule and the most efficient port usage.
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.1
 * @see com.cyara.apisamples.EligibleCampaignRun
 * @see com.cyara.apisamples.EligibleTestCase
 */
public class SchedulePulseCampaigns {
	//Change to your Pulse test cases folder
	static String pulseFolder = "CHANGE_ME";

	//Change to your Pulse Plan ID
	static int pulsePlanId = -1;

	public static void main(String[] argv) {
		ApiFunctionalities connInfo = new ApiFunctionalities();

		JsonObject planInfo = connInfo.getPlanDetails(pulsePlanId);
		int planConcurrency = planInfo.getInt("concurrency");

		SimpleDateFormat trackedFormat = new SimpleDateFormat("EEEE HH:mm yyyy");

		/*
		 * Using 1995 because that year started on a Sunday.
		 * This becomes necessary when comparing days coming
		 * before or after other days when processing dates
		 * in Java.
		 */
		String endOfWeekTime = "Saturday 23:59 1995";
		Date endOfWeekDate = null;
		try {
			endOfWeekDate = trackedFormat.parse(endOfWeekTime);
		} catch (ParseException e) {
			e.printStackTrace();
			System.exit(1);
		}

		ArrayList<Integer> pulseTestCaseIds = getPulseTestCases(connInfo);
		ArrayList<String> pulseTestCaseXmls = new ArrayList<>();

		//Export and store all pulse test cases to be used throughout program
		for (int tcIndex = 0; tcIndex < pulseTestCaseIds.size(); tcIndex++) {
			String tcXml = connInfo.exportTestCase(pulseTestCaseIds.get(tcIndex), "IncludeBlocks");
			Document testCaseDoc = connInfo.convertStringToDocument(tcXml);
			testCaseDoc.getDocumentElement().normalize();

			String tcDesc = testCaseDoc.getElementsByTagName("Description").item(0).getTextContent();

			//The description must have braces, indicating the json format presence
			if (tcDesc.contains("{") && tcDesc.contains("}")) {
				pulseTestCaseXmls.add(tcXml);
			}
			else {
				//This test case didn't have a json desc, remove it from the list
				pulseTestCaseIds.remove(tcIndex);
				tcIndex--;
			}
		}

		//Store all the unique frequencies desired by each test case, collectively
		ArrayList<Integer> uniqueFrequencies = getUniqueFrequencies(pulseTestCaseXmls, connInfo);
		Collections.sort(uniqueFrequencies);

		int eligibleRunId = 0;
		ArrayList<EligibleCampaignRun> allRuns = new ArrayList<EligibleCampaignRun>();

		/*
		 * For every frequency, grab all the runs.
		 * 
		 * A run is defined as a set of test cases with the same
		 * frequency that have a scheduled run time in their
		 * description json that overlaps with the time of week
		 * we're currently processing as we move through the
		 * week.
		 */
		for (int freq = 0; freq < uniqueFrequencies.size(); freq++) {
			int desiredFreq = uniqueFrequencies.get(freq);

			//Use 1995 because that year started on a Sunday
			String currentTime = "Sunday 00:00 1995";
			Date currentDate = null;
			try {
				currentDate = trackedFormat.parse(currentTime);
			} catch (ParseException e) {
				e.printStackTrace();
				System.exit(1);
			}



			System.out.println("STARTING THE RUN CREATIONS FOR THIS FREQUENCY: " + desiredFreq); 

			/*
			 * parent loop here checking if the currentTime is at Saturday 23:59,
			 * meaning we're at the end of the week and thus done with this timeframe
			 */
			while (!currentDate.equals(endOfWeekDate)) {
				System.out.println("At the top of loop for this frequency building runs thru week");
				Date campaignRunStartDate;
				Date campaignRunEndDate;

				//newEligibleDate is just used as the workhorse to find the next relevant time to what we're processing
				Date newEligibleDate = endOfWeekDate;

				/*
				 * Here we find the next startTime or endTime and assign it to currentTime
				 * This will be used to determine when this campaign run should start
				 */
				for (int tcIndex = 0; tcIndex < pulseTestCaseIds.size(); tcIndex++) {

					//Convert this test case xml String into a Document
					Document testCaseDoc = connInfo.convertStringToDocument(pulseTestCaseXmls.get(tcIndex));
					testCaseDoc.getDocumentElement().normalize();

					//Get Description and convert to JsonObject
					String tcDesc = testCaseDoc.getElementsByTagName("Description").item(0).getTextContent();
					tcDesc = tcDesc.substring(tcDesc.indexOf("{"), tcDesc.lastIndexOf("}")+1);

					JsonObject tcScheduleInfo = connInfo.convertAPIOutputToJson(tcDesc);

					int testCaseFreq = tcScheduleInfo.getInt("frequency");

					/*
					 * If this test case frequency matches the frequency we're
					 * working with, check if it's the next to start or end
					 */
					if (testCaseFreq == desiredFreq) {
						JsonArray tcSchedule = tcScheduleInfo.getJsonArray("schedule");

						//Loop thru all the description json schedules for this test case
						for (int scheduleIndex = 0; scheduleIndex < tcSchedule.size(); scheduleIndex++) {
							JsonObject eligibleEntry = tcSchedule.getJsonObject(scheduleIndex);
							String day = eligibleEntry.getString("day");
							String startTime = eligibleEntry.getString("startTime");
							String endTime = eligibleEntry.getString("endTime");

							Date startDate = null;
							Date endDate = null;
							try {
								startDate = trackedFormat.parse(day + " " + startTime + " 1995");
								endDate = trackedFormat.parse(day + " " + endTime + " 1995");
							} catch (ParseException e) {
								e.printStackTrace();
								System.exit(1);
							}

							/*
							 * newEligibleDate was assigned Saturday at 11:59 before this loop's loop.
							 * currentDate is the start of the week or, if we've already create at
							 * least one run, it's +1 minute after the run we just created for this frequency ended
							 * Because we add +1 minute to currentDate at the end of the last generation,
							 * that's why we check for equality to currendDate, because currentDate
							 * hasn't been checked and won't overlap with anything.
							 * 
							 * The first if will only ever get triggered once we've had at least
							 * one run generated and now we've looped back up here. This is because we're
							 * checking if this test case's scheduled start is before the current date, which
							 * is impossible for the first generation because currentDate is assigned Sunday 00:00.
							 * We're also checking if this run ends after currentDate, implying the previous generation
							 * ended before this test case ends, even while this test case is running (and we'll
							 * want to have this campaign run start at currentDate so it will at least capture
							 * the remaining time for this test case. Finally, we're checking if currentDate
							 * is before newEligibleDate, because we always and only want to to earlier.
							 * 
							 * For the second if, we're checking if this test case's startDate is before newEligibleDate
							 * which was set to EoW, and if it's equal to or later than currentDate, to grab
							 * the next earliest start time for this run.
							 * .
							 * Together, these us back earlier and earlier, but not too early, to the time
							 * this campaign run should start.
							 */
							if (startDate.before(currentDate)
									&& (endDate.after(currentDate) || endDate.equals(currentDate))
									&& currentDate.before(newEligibleDate)) {
								newEligibleDate = currentDate;
							}
							if (startDate.before(newEligibleDate)
									&& (startDate.after(currentDate) || startDate.equals(currentDate))) {
								newEligibleDate = startDate;
							}
						}
					}
				}

				campaignRunStartDate = newEligibleDate;
				currentDate = campaignRunStartDate;

				System.out.println("campaignRunStartDate: " + campaignRunStartDate);


				ArrayList<EligibleTestCase> eligibleTestCasesForCurTimeAndFreq = new ArrayList<>();

				/*
				 * Loop through every TC to find eligible ones for the current time
				 * These eligible runs get stored into eligibleTestCasesForCurTimeAndFreq
				 * Eligibility is determined if the frequency is the same, and there is
				 * a schedule going during the currentDate
				 */
				for (int tcIndex = 0; tcIndex < pulseTestCaseIds.size(); tcIndex++) {
					int tcId = pulseTestCaseIds.get(tcIndex);

					Document testCaseDoc = connInfo.convertStringToDocument(pulseTestCaseXmls.get(tcIndex));
					testCaseDoc.getDocumentElement().normalize();

					NodeList scenarioList = testCaseDoc.getElementsByTagName("Scenario");
					int scenarios = scenarioList.getLength();
					System.out.println(scenarios);
					if (scenarios == 0) {
						scenarios = 1;
					}

					System.out.println("Root element :" + testCaseDoc.getDocumentElement().getNodeName());
					String tcDesc = testCaseDoc.getElementsByTagName("Description").item(0).getTextContent();
					tcDesc = tcDesc.substring(tcDesc.indexOf("{"), tcDesc.lastIndexOf("}")+1);
					System.out.println(tcDesc);

					JsonObject tcScheduleInfo = connInfo.convertAPIOutputToJson(tcDesc);

					int testCaseFreq = tcScheduleInfo.getInt("frequency");

					if (testCaseFreq == desiredFreq) {

						JsonArray tcSchedule = tcScheduleInfo.getJsonArray("schedule");

						//Loop thru every schedule for this test case, checking if a run is happening during currentDate
						for (int scheduleIndex = 0; scheduleIndex < tcSchedule.size(); scheduleIndex++) {
							JsonObject eligibleEntry = tcSchedule.getJsonObject(scheduleIndex);
							String day = eligibleEntry.getString("day");
							String startTime = eligibleEntry.getString("startTime");
							String endTime = eligibleEntry.getString("endTime");

							Date startDate = null;
							Date endDate = null;
							try {
								startDate = trackedFormat.parse(day + " " + startTime + " 1995");
								endDate = trackedFormat.parse(day + " " + endTime + " 1995");
							} catch (ParseException e) {
								e.printStackTrace();
								System.exit(1);
							}

							/*
							 * Check if the current date is at or after the start but
							 * before the end. We don't want at the end or else this
							 * could get false failures running just after eligible time
							 * ends.
							 * 
							 * In other words, check if the currentDate is within this run time
							 */
							if ((currentDate.equals(startDate) || currentDate.after(startDate)) && currentDate.before(endDate)) {
								System.out.println("ADDING " + tcId);
								eligibleTestCasesForCurTimeAndFreq.add(new EligibleTestCase(tcId, tcScheduleInfo.getInt("minuteDuration"), testCaseFreq, scenarios, day, startTime, endTime, startDate, endDate));

								//We found an eligible time, so we're done processing this test case
								break;
							}
						}

					}

				}

				System.out.println("Finished looping TCs to find eligible ones for the current time with size: " + eligibleTestCasesForCurTimeAndFreq.size() + " currentDate: " + currentDate);

				/*
				 * We've now grabbed all the eligible test cases
				 * for this frequency. Let's now grab the next start
				 * or end time to become to new eligible time. This then
				 * makes the current eligible range curEligible to newEligible
				 * AKA here is where we find the endTime and assign it to currentDate
				 * We'll subtract 1 from endTime if we used the next start as an end trigger
				 */

				newEligibleDate = endOfWeekDate;

				boolean usedStartDate = false;

				for (int tcIndex = 0; tcIndex < pulseTestCaseIds.size(); tcIndex++) {

					Document testCaseDoc = connInfo.convertStringToDocument(pulseTestCaseXmls.get(tcIndex));
					testCaseDoc.getDocumentElement().normalize();

					String tcDesc = testCaseDoc.getElementsByTagName("Description").item(0).getTextContent();
					tcDesc = tcDesc.substring(tcDesc.indexOf("{"), tcDesc.lastIndexOf("}")+1);
					System.out.println(tcDesc);

					JsonObject tcScheduleInfo = connInfo.convertAPIOutputToJson(tcDesc);

					int testCaseFreq = tcScheduleInfo.getInt("frequency");

					if (testCaseFreq == desiredFreq) {
						System.out.println("Test Case " + pulseTestCaseIds.get(tcIndex) + " is matching our freq");
						JsonArray tcSchedule = tcScheduleInfo.getJsonArray("schedule");

						for (int scheduleIndex = 0; scheduleIndex < tcSchedule.size(); scheduleIndex++) {
							JsonObject eligibleEntry = tcSchedule.getJsonObject(scheduleIndex);
							String day = eligibleEntry.getString("day");
							String startTime = eligibleEntry.getString("startTime");
							String endTime = eligibleEntry.getString("endTime");

							Date startDate = null;
							Date endDate = null;
							try {
								startDate = trackedFormat.parse(day + " " + startTime + " 1995");
								endDate = trackedFormat.parse(day + " " + endTime + " 1995");
							} catch (ParseException e) {
								e.printStackTrace();
								System.exit(1);
							}

							/*
							 * newEligibleDate was assigned Saturday at 11:59 before this loop's loop.
							 * So we're now checking for earlier and earlier dates that are still after
							 * currentDate (which is currently set to this run's start time) but
							 * before newEligibleDate. This will yield us when the campaign should end.
							 */
							if (startDate.before(newEligibleDate) && startDate.after(currentDate)) {
								newEligibleDate = startDate;
								usedStartDate = true;
								System.out.println("Assigning newEligibleDate to startDate: " + startDate);
							}
							else if (endDate.before(newEligibleDate) && endDate.after(currentDate)) {
								newEligibleDate = endDate;
								usedStartDate = false;
								System.out.println("Assigning newEligibleDate to endDate: " + endDate);
							}
						}
					}
				}

				/*
				 * We found the earliest end time for this frequency, subtract 1 from it and store but
				 * only subtract if our newEligibleDate was assigned to a startDate, because
				 * the endDate in description json already has 1 subtracted from when things break
				 */
				if (!newEligibleDate.equals(endOfWeekDate) && usedStartDate) {
					java.util.Calendar morphCal = java.util.Calendar.getInstance();
					morphCal.setTime(newEligibleDate);
					morphCal.add(java.util.Calendar.MINUTE, -1);  // number of minutes to subtract
					newEligibleDate = morphCal.getTime();
				}

				currentDate = newEligibleDate;
				campaignRunEndDate = newEligibleDate;

				System.out.println("campaignRunEndDate: " + campaignRunEndDate);


				if (!eligibleTestCasesForCurTimeAndFreq.isEmpty()) {
					/*
					 * Sort the test cases in reverse order. This ensures the fastest possible runtime calculated in getReqPortsForFreq.
					 * By going in reverse, we run the longest first. This takes advantage of multi-port accounts whereby
					 * one port can be running a huge test case while other ports are running the increasingly short test cases
					 */
					Collections.sort(eligibleTestCasesForCurTimeAndFreq, Comparator.comparingInt(EligibleTestCase::getMinuteDuration));
					Collections.reverse(eligibleTestCasesForCurTimeAndFreq);

					int[] runDetails = getReqPortsForFreq(eligibleTestCasesForCurTimeAndFreq, desiredFreq);

					int requiredPortsForFrequencyRun = runDetails[0];

					if (requiredPortsForFrequencyRun > planConcurrency) {
						System.err.println("ERROR: The test cases for the " + desiredFreq + " frequency require "
								+ requiredPortsForFrequencyRun + " ports to finish on time. This is more than the "
								+ planConcurrency + " port(s) your plan is licensed for.");
						System.exit(1);
					}
					else {
						System.out.println("You will need " + requiredPortsForFrequencyRun + " port(s) for this specific set of " + desiredFreq + " minute frequency test case(s)");
					}

					int runDuration = runDetails[1];

					/*
					 * Calculate the number of minutes this scheduled run is,
					 * and determine if it can run more than once in that time
					 */
					long startToEndScheduleDiff = campaignRunEndDate.getTime() - campaignRunStartDate.getTime();
					int startToEndScheduleMinutes = (int) (startToEndScheduleDiff / (60 * 1000.0));

					boolean runsMultipleTimes = (startToEndScheduleMinutes / runDuration) > 0;

					/*
					 * This Date only gets used if this run only executes
					 * once in the schedule, such as if this run is
					 * a Regression Monitoring suite
					 */
					Date actualOneAndDoneEndDate = new Date();

					/*
					 * This scheduled run will fire off more than once,
					 * so we need to change the campaignRunEndDate to
					 * account for the buffer time needed to prevent
					 * the campaign run from going over its schedule
					 */
					if (runsMultipleTimes) {
						int runCountsPreChange = (startToEndScheduleMinutes / runDuration) + 1;

						java.util.Calendar morphCal = java.util.Calendar.getInstance();
						morphCal.setTime(campaignRunEndDate);
						morphCal.add(java.util.Calendar.MINUTE, runDuration * -1);  // number of minutes to subtract
						campaignRunEndDate = morphCal.getTime();

						startToEndScheduleDiff = campaignRunEndDate.getTime() - campaignRunStartDate.getTime();
						startToEndScheduleMinutes = (int) (startToEndScheduleDiff / (60 * 1000.0));

						int runCountsPostChange = (startToEndScheduleMinutes / runDuration) + 1;

						if (runCountsPreChange != runCountsPostChange) {
							System.err.println("WARNING: The number of runs for the " + desiredFreq + " minute frequency run from "
									+ campaignRunStartDate + " to " + campaignRunEndDate + "\n    went from " + runCountsPreChange +
									" to " + runCountsPostChange + " after factoring in a buffer time of " + runDuration + " minute(s).");
						}
					}
					else {
						/*
						 * If the run is a one and done, we need to calculate
						 * when it actually ends because the campaignRunEndDate
						 * isn't relevant for this but we still need to track
						 * how many ports this run will be using and for how long
						 * to, later in the code, check for breaches in port usage
						 */
						System.out.println("This run is a one and done, so let's find and store when it will ACTUALLY end");
						//This run is a one and done, like a Regression Monitoring Suite, so let's find and store when it will ACTUALLY end
						java.util.Calendar morphCal = java.util.Calendar.getInstance();
						morphCal.setTime(campaignRunStartDate);
						morphCal.add(java.util.Calendar.MINUTE, runDuration);  // number of minutes to subtract
						actualOneAndDoneEndDate = morphCal.getTime();
					}

					System.out.println("End time after buffer applied (if this run launches multiple times which is " + runsMultipleTimes + "): " + campaignRunEndDate);

					allRuns.add(new EligibleCampaignRun(campaignRunStartDate, campaignRunEndDate, eligibleTestCasesForCurTimeAndFreq, desiredFreq, requiredPortsForFrequencyRun, runDuration, runsMultipleTimes, actualOneAndDoneEndDate, eligibleRunId));

					eligibleRunId++;

					//Increase currentDate by 1 to force the top of loop to look at the next run times
					if (!currentDate.equals(endOfWeekDate)) {
						java.util.Calendar morphCal = java.util.Calendar.getInstance();
						morphCal.setTime(currentDate);
						morphCal.add(java.util.Calendar.MINUTE, 1);  // number of minutes to subtract
						currentDate = morphCal.getTime();
					}
				}

			} //End of loop for this frequency going thru week

		} //End of loop going thru all frequencies

		/*
		 * We've now grabbed all the runs that will occur
		 * for every timeframe. From there, we can loop
		 * thru each run and check if there are any campaigns
		 * that are prebuilt for that set of test cases, frequency, and ports.
		 * This requires checking for any simultaneous runs and
		 * calculating the ports and time required.
		 * 
		 * The frequencies are already sorted because our loop way above
		 * hunting for test case freq was sorted lowest to highest freq
		 */

		//Loop to check if we ever use more ports combined than we're licensed for
		findPortBreaches(allRuns, planConcurrency);

		//Create a date in the past so the campaign runs immediately if it can when enabled
		String requestedScheduledRunDate = DateTimeFormatter.ISO_INSTANT.format(Instant.now().minus(1, ChronoUnit.DAYS));

		SimpleDateFormat scheduleDayFormat = new SimpleDateFormat("EEEE");
		SimpleDateFormat scheduleFromToFormat = new SimpleDateFormat("HH:mm:ss");

		//here is where we combine our runs if they have the same frequency, test cases, and ports
		for (int x = 0; x < allRuns.size(); x++) {
			EligibleCampaignRun currRun = allRuns.get(x);
			int currRunId = currRun.getId();
			int currFrequency = currRun.getCampaignFrequency();

			String runTimeJson = "{\r\n" + 
					"				\"day\":\"" + scheduleDayFormat.format(currRun.getStartDate()) + "\",\r\n" + 
					"				\"startTime\":\"" + scheduleFromToFormat.format(currRun.getStartDate()) + "\",\r\n" + 
					"				\"endTime\":\"" + scheduleFromToFormat.format(currRun.getEndDate()) + "\"\r\n" + 
					"			},";

			ArrayList<Integer> currTestCaseIds = new ArrayList<Integer>();

			ArrayList<EligibleTestCase> currTestCases = currRun.getEligibleTestCases();

			String testCasesJson = "";

			//Add every test case in this run to the test cases json needed for campaign creation
			for (int currTcIndex = 0; currTcIndex < currTestCases.size(); currTcIndex++) {
				currTestCaseIds.add(currTestCases.get(currTcIndex).getTestCaseId());

				testCasesJson = testCasesJson + "{\r\n" + 
						"			\"testCaseId\":" + currTestCases.get(currTcIndex).getTestCaseId() + ",\r\n" + 
						"			\"probability\":100,\r\n" + 
						"			\"channel\":\"Voice\"\r\n" + 
						"		},";
			}

			//Sort the list of current test case Ids, so we can compare to the other runs
			Collections.sort(currTestCaseIds);

			for (int y = 0; y < allRuns.size(); y++) {
				EligibleCampaignRun checkedRun = allRuns.get(y);
				int checkedRunId = checkedRun.getId();
				int checkedFrequency = checkedRun.getCampaignFrequency();

				/*
				 * This checked run is a different run but has the same frequency.
				 * Let's explore if it has the same test cases and thus can
				 * be added to testCasesJson
				 */
				if (currRunId != checkedRunId && currFrequency == checkedFrequency) {
					ArrayList<EligibleTestCase> checkedTestCases = checkedRun.getEligibleTestCases();

					ArrayList<Integer> checkedTestCaseIds = new ArrayList<Integer>();

					for (int checkedTcIndex = 0; checkedTcIndex < checkedTestCases.size(); checkedTcIndex++) {
						checkedTestCaseIds.add(checkedTestCases.get(checkedTcIndex).getTestCaseId());
					}

					Collections.sort(checkedTestCaseIds);

					//These two runs have the same test cases (on top of same freq). Let's add the schedule and remove this run
					if (checkedTestCaseIds.equals(currTestCaseIds)) {

						runTimeJson = runTimeJson + "{\r\n" + 
								"				\"day\":\"" + scheduleDayFormat.format(checkedRun.getStartDate()) + "\",\r\n" + 
								"				\"startTime\":\"" + scheduleFromToFormat.format(checkedRun.getStartDate()) + "\",\r\n" + 
								"				\"endTime\":\"" + scheduleFromToFormat.format(checkedRun.getEndDate()) + "\"\r\n" + 
								"			},";

						allRuns.remove(y);
						y--;
					}
				}
			}

			//Remove the extra comma from each json String we built
			runTimeJson = runTimeJson.substring(0, runTimeJson.length() - 1);
			testCasesJson = testCasesJson.substring(0, testCasesJson.length() - 1);

			//We now all the pieces for a campaign, assemble and build out the json for the API to build it

			String campaignName = "Index: " + currRunId + " | Ports: " + currRun.getRequiredPortsForFrequencyRun() + " | Freq: " + currFrequency;

			String campaignBody = "{\r\n" + 
					"	\"active\":false,\r\n" + 
					"	\"name\":\"" + campaignName + "\",\r\n" + 
					"	\"maximumAps\":1.0,\r\n" + 
					"	\"frequency\": \"00:" + currFrequency + ":00\",\r\n" +
					"	\"concurrency\":" + currRun.getRequiredPortsForFrequencyRun() + ",\r\n" +
					"	\"requestedRunDate\":\"" + requestedScheduledRunDate + "\",\r\n" + 
					"	\"scheduledRunDate\":\"" + requestedScheduledRunDate + "\",\r\n" + 
					"	\"channel\":\"Voice\",\r\n" + 
					"	\"testCaseDistributionProfile\":\"RoundRobin\",\r\n" + 
					"	\"schedules\":[\r\n" + 
					runTimeJson +
					"	],\r\n" + 
					"	\"schedulePauses\":[],\r\n" + 
					"	\"planId\":" + pulsePlanId + ",\r\n" + 
					"	\"planType\":\"Pulse\",\r\n" + 
					"	\"testCases\":[\r\n" + 
					testCasesJson +
					"	]\r\n" + 
					"}";

			//Create the campaign
			JsonObject campaignCreation = connInfo.createCampaign(campaignBody);
			System.out.println(campaignCreation);
		}

	}


	/**
	 * Loops through all the campaign runs and checks for any
	 * combination of runs taking place simultaneously that
	 * breach the Pulse plan's port count
	 * 
	 * @param allRuns All the campaign runs (not in campaigns, just the schedule, test cases, frequency, ports, etc.) 
	 * @param planConcurrency The number of concurrent calls the Cyara plan supports
	 * @since 1.0
	 */
	private static void findPortBreaches(ArrayList<EligibleCampaignRun> allRuns, int planConcurrency) {

		//The main loop to check each run's maximum simultaneous port concurrency
		for (int x = 0; x < allRuns.size(); x++) {
			//Get this run, which is what we'll be checking all the other runs against for simultaneity
			EligibleCampaignRun currRun = allRuns.get(x);
			int currRunId = currRun.getId();
			Date currRunStartDate = currRun.getStartDate();
			Date currRunEndDate = currRun.getEndDate();

			//If this run isn't multiple times but just a one and done, get the true end time
			if (!currRun.getRunsMultipleTimes()) {
				currRunEndDate = currRun.getActualOneAndDoneEndDate();
			}

			int currRunPorts = currRun.getRequiredPortsForFrequencyRun();

			//ArrayList for run that is, at some point, overlapping
			ArrayList<EligibleCampaignRun> overlappingRuns = new ArrayList<EligibleCampaignRun>();

			//Loop thru all the other runs, checking for any overlap to currRun, and store ones that do
			for (int y = 0; y < allRuns.size(); y++) {
				EligibleCampaignRun checkedRun = allRuns.get(y);
				int checkedRunId = checkedRun.getId();

				if (currRunId != checkedRunId) {
					Date checkedStartDate = checkedRun.getStartDate();
					Date checkedEndDate = checkedRun.getEndDate();

					System.out.println("checkedStartDate " + checkedStartDate);
					System.out.println("checkedEndDate " + checkedEndDate);

					//Check if this run isn't just a one and done, but an actual run
					if (!checkedRun.getRunsMultipleTimes()) {
						checkedEndDate = checkedRun.getActualOneAndDoneEndDate();
					}

					//Check if these campaign runs overlap
					if ((checkedStartDate.equals(currRunStartDate) || checkedEndDate.equals(currRunEndDate) || checkedStartDate.equals(currRunEndDate) || checkedEndDate.equals(currRunEndDate)) ||
							(checkedStartDate.after(currRunStartDate) && checkedStartDate.before(currRunEndDate)) ||
							(checkedEndDate.after(currRunStartDate) && checkedEndDate.before(currRunEndDate))) {
						System.out.println("Adding to run ID " + currRunId + " the following simultaneous run to check: " + checkedRunId);

						overlappingRuns.add(checkedRun);
					}
				}
			}

			//We now have all the runs that overlap at some point, let's see if any breach
			for (int y = 0; y < overlappingRuns.size(); y++) {
				//The checked run is somewhere inside the current run
				EligibleCampaignRun overlappingRun = overlappingRuns.get(y);
				int overlappedRunId = overlappingRun.getId();

				Date overlappedStartDate = overlappingRun.getStartDate();
				Date overlappedEndDate = overlappingRun.getEndDate();

				if (!overlappingRun.getRunsMultipleTimes()) {
					overlappedEndDate = overlappingRun.getActualOneAndDoneEndDate();
				}

				//"Trim" the leading and trailing run times beyond curr Run. We don't care about them
				if (overlappedStartDate.before(currRunStartDate)) {
					overlappedStartDate = currRunStartDate;
				}
				if (overlappedEndDate.after(currRunEndDate)) {
					overlappedEndDate = currRunEndDate;
				}

				//Simultaneous variables. Used to combine overlapped run and compared runs 
				int simultaneousPorts = overlappingRun.getRequiredPortsForFrequencyRun();
				ArrayList<EligibleCampaignRun> simultaneousRuns = new ArrayList<EligibleCampaignRun>();
				simultaneousRuns.add(overlappingRun);

				/*
				 * For this specific run that overlaps currRun, we're now going to
				 * loop through all the other overlapping runs and
				 * find and grab any that overlap this specific overlap.
				 * This will tell us the most simultaneous port usage.
				 * Current Run Ports
				 * +
				 * Overlapping (trimmed to currRun start and stop) Run Ports
				 * +
				 * Runs Overlapping (trimmed to specific overlap Run start and stop) the Overlap Ports
				 */
				for (int z = 0; z < overlappingRuns.size(); z++) {
					//The compared run is where we compare it to the overlapping run to find the simultaneity, which is done for all other runs besides overlapped
					EligibleCampaignRun comparedRun = overlappingRuns.get(z);
					int comparedRunId = comparedRun.getId();

					if (overlappedRunId != comparedRunId) {

						Date comparedStartDate = comparedRun.getStartDate();
						Date comparedEndDate = comparedRun.getEndDate();

						if (!comparedRun.getRunsMultipleTimes()) {
							comparedEndDate = comparedRun.getActualOneAndDoneEndDate();
						}

						//If the comparison starts before but end on or beyond the overlapped start, trim start for simplicity
						if (comparedStartDate.before(overlappedStartDate) && (comparedEndDate.equals(overlappedStartDate) || comparedEndDate.after(overlappedStartDate))) {
							comparedStartDate = overlappedStartDate;
						}

						//If the comparison ends after but starts on or before the check end, trim for simplicity
						if (comparedEndDate.after(overlappedEndDate) && (comparedStartDate.equals(overlappedEndDate) || comparedStartDate.before(overlappedEndDate))) {
							comparedEndDate = overlappedEndDate;
						}

						//This compared run does, at some point, run with the overlapped run (which is overlapping with curr run)
						if ((comparedStartDate.equals(overlappedStartDate) || comparedEndDate.equals(overlappedStartDate) ||
								comparedStartDate.equals(overlappedEndDate) || comparedEndDate.equals(overlappedEndDate))
								||
								(comparedStartDate.after(overlappedStartDate) && comparedStartDate.before(overlappedEndDate))
								||
								(comparedEndDate.after(overlappedStartDate) && comparedEndDate.before(overlappedEndDate))) {
							System.out.println("Stacking to run id " + overlappedRunId + "'s port usge the following compared run ports: " + comparedRun.getRequiredPortsForFrequencyRun());

							simultaneousPorts += comparedRun.getRequiredPortsForFrequencyRun();

							simultaneousRuns.add(comparedRun);
						}
					}
				}

				//We just grabbed all the simultaneous runs for this current run, check if we breached our max ports
				if (currRunPorts + simultaneousPorts > planConcurrency) {

					simultaneousRuns.add(0, currRun);

					System.err.println("ERROR: The following runs occurring together breach your plan's ports. "
							+ "Please reduce your number of test cases, their run times, or increase your frequency or "
							+ "your licensed number of ports. Details are below:");

					System.err.println((currRunPorts + simultaneousPorts) + " ports used. Only " + planConcurrency + " port(s) licensed.");

					for (int run = 0; run < simultaneousRuns.size(); run++) {
						System.err.println(simultaneousRuns.get(run).toString());
					}
					System.exit(1);
				}
				else {
					System.out.println("Highest combination for curr run " + currRunId + " simultaneous with overlap run " + overlappedRunId + " is only " + (currRunPorts + simultaneousPorts) + " port(s).");
				}
			}

		}
	}


	/**
	 * Loops through all the eligible test cases that will become a campaign run
	 * and checks for duration errors, calculates how many ports these
	 * test cases will use, how long the run will take, and how many calls will be made.
	 * 
	 * @param eligibleTestCasesForCurTimeAndFreq All the test cases that are running
	 * during the current time when this method is called with the same frequency as desired
	 * @param desiredFreq The frequency that was in the desc json for these test cases
	 * @return An int array containing the least number of ports required, the
	 * run duration, and the total number of calls for this run
	 * @since 1.0
	 */
	private static int[] getReqPortsForFreq(ArrayList<EligibleTestCase> eligibleTestCasesForCurTimeAndFreq, int desiredFreq) {
		int[] runInfo = new int[3];
		ArrayList<EligibleTestCase> eligibleTCsCopy = new ArrayList<>(eligibleTestCasesForCurTimeAndFreq);

		int totalScenarios = 0;

		for (int tcIndex = 0; tcIndex < eligibleTCsCopy.size(); tcIndex++) {
			EligibleTestCase eligibleTC = eligibleTCsCopy.get(tcIndex);
			totalScenarios += eligibleTC.getScenarios();
			int tcDur = eligibleTC.getMinuteDuration();

			if (tcDur >= desiredFreq) {
				System.err.println("ERROR: The following test case is equal to or longer than "
						+ "your desired frequency, thus making this frequency mathematically impossible: " + eligibleTC.getTestCaseId());
				System.err.println("Desired Frequency: " + desiredFreq);
				System.err.println("Test Case Minute Duration: " + tcDur);
				System.exit(1);
			}
		}

		/*
		 * Loop thru each simulated number of ports until we find the number needed to finish less than our desiredFreq.
		 * The maximum number of necessary ports is the same as totalScenarios, because this would simulate 1 port per call,
		 * and any more ports than that would be unused and completely irrelevant
		 */
		for (int attemptedPorts = 1; attemptedPorts <= totalScenarios; attemptedPorts++) {
			eligibleTCsCopy = new ArrayList<>(eligibleTestCasesForCurTimeAndFreq);
			System.out.println("Using ports " + attemptedPorts);
			int[] portDurations = new int[attemptedPorts];

			boolean tcAdded = true;

			//Parent loop controller to grab all scenario durations
			while (tcAdded) {
				tcAdded = false;

				/*
				 * Loop thru all the test cases over and over until
				 * we've grabbed and stored all the scenario durations
				 */
				for (int tc = 0; tc < eligibleTCsCopy.size(); tc++) {
					EligibleTestCase eligibleTC = eligibleTCsCopy.get(tc);
					EligibleTestCase dupeTC = new EligibleTestCase(eligibleTC.getTestCaseId(), eligibleTC.getMinuteDuration(), eligibleTC.getTestCaseFrequency(),
							eligibleTC.getScenarios(), eligibleTC.getDay(), eligibleTC.getStartTime(), eligibleTC.getEndTime(), eligibleTC.getTestCaseStartDate(),
							eligibleTC.getTestCaseEndDate());
					int curTCScenarios = dupeTC.getScenarios();
					System.out.println("curTCScenarios: " + curTCScenarios);
					if (curTCScenarios > 0) {
						tcAdded = true;
						curTCScenarios = curTCScenarios - 1;
						dupeTC.setScenarios(curTCScenarios);
						eligibleTCsCopy.set(tc, dupeTC);

						int shortestPortIndex = 0;
						int shortestTrackedTime = portDurations[shortestPortIndex];

						for (int port = 1; port < portDurations.length; port++) {
							int portDuration = portDurations[port];

							if (portDuration < shortestTrackedTime) {
								shortestPortIndex = port;
								shortestTrackedTime = portDuration;
							}
						}

						/*
						 * We've now found the port with the current shortest duration.
						 * Let's add this test case's scenario duration to it.
						 */

						System.out.println("Adding to portDurations index " + shortestPortIndex + " a "
								+ "test case with a " + eligibleTC.getMinuteDuration() + " minute(s) duration to the "
								+ "port's current duration of " + shortestTrackedTime);
						portDurations[shortestPortIndex] = shortestTrackedTime + eligibleTC.getMinuteDuration();
					}

				}
			}

			/*
			 * We've now grabbed all the scenario durations for this port count.
			 * Find and grab the longest of those durations.
			 * If the longest of those durations is less than our frequency,
			 * then we've found the minimum ports necessary.
			 */

			int longestDuration = portDurations[0];
			for (int port = 1; port < portDurations.length; port++) {
				if (portDurations[port] > longestDuration) {
					longestDuration = portDurations[port];
				}
			}

			if (longestDuration < desiredFreq) {
				System.out.println("For your " + desiredFreq + " minute frequency, the longest duration for a given port using " + attemptedPorts + " ports is " + longestDuration);
				runInfo[0] = attemptedPorts;
				runInfo[1] = longestDuration;
				runInfo[2] = totalScenarios;
				return runInfo;
			}

		}

		return runInfo;
	}


	/**
	 * Loops through every tracked test case, storing every unique
	 * "frequency" field in the description json
	 * 
	 * @param pulseTestCaseXmls The XML of every tracked test case
	 * @param connInfo The controller to the API credentials and methods
	 * @return An Integer ArrayList with every unique frequency
	 * @since 1.0
	 */
	private static ArrayList<Integer> getUniqueFrequencies(ArrayList<String> pulseTestCaseXmls, ApiFunctionalities connInfo) {
		ArrayList<Integer> uniqueFrequencies = new ArrayList<>();

		//Loop through every test case
		for (int tcIndex = 0; tcIndex < pulseTestCaseXmls.size(); tcIndex++) {

			//Convert the Cyara XML of the test case to a Document
			Document testCaseDoc = connInfo.convertStringToDocument(pulseTestCaseXmls.get(tcIndex));
			testCaseDoc.getDocumentElement().normalize();

			//Grab the description, then just the open and closing braces of the json
			String tcDesc = testCaseDoc.getElementsByTagName("Description").item(0).getTextContent();
			tcDesc = tcDesc.substring(tcDesc.indexOf("{"), tcDesc.lastIndexOf("}")+1);

			//Convert our description json to a JsonObject to parse
			JsonObject tcScheduleInfo = connInfo.convertAPIOutputToJson(tcDesc);

			//Get the user-specified frequency
			int frequency = tcScheduleInfo.getInt("frequency");

			//If this frequency is new, store it
			if (!uniqueFrequencies.contains(frequency)) {
				uniqueFrequencies.add(frequency);
			}
		}

		return uniqueFrequencies;
	}


	/**
	 * Loops through all the test case on the account, making
	 * the API request to list test cases until we've
	 * capture and stored all of them.
	 * 
	 * @param connInfo The controller to the API credentials and methods
	 * @return The ArrayList of every test case on the account
	 * @since 1.0
	 */
	private static ArrayList<Integer> getPulseTestCases(ApiFunctionalities connInfo) {
		ArrayList<Integer> testCaseIds = new ArrayList<>();

		boolean morePages = true;
		String nextPage = "";
		JsonObject apiObject = connInfo.listTestCases();

		while (morePages) {
			if (apiObject.containsKey("nextPage")) {
				if (apiObject.isNull("nextPage")) {
					morePages = false;
				}
				else {
					nextPage = apiObject.getString("nextPage");
				}

				JsonArray resultsArr = apiObject.getJsonArray("results");
				for (int resultIndex = 0; resultIndex < resultsArr.size(); resultIndex++) {
					//Get specific test case
					JsonObject testCase = resultsArr.getJsonObject(resultIndex);

					int testCaseId = testCase.getInt("testCaseId");

					String folderPath = testCase.getString("folderPath");

					//If it's in our folder, store its ID
					if(folderPath.startsWith(pulseFolder)) {
						testCaseIds.add(testCaseId);
					}
				}
			}
			else { //nextPage doesn't even exist, something is totally wrong
				morePages = false;
			}

			if (morePages) { //After processing that call, do we still have pages? Then make the next call here
				apiObject = connInfo.convertAPIOutputToJson(connInfo.makeApiGet(nextPage));
			}
		}

		return testCaseIds;
	}
}
