package com.cyara.utilities;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.ArrayList;

/**
 * Class built for processing the failures found in
 * MatchedConnectionFailures, primarily to generate
 * the Json needed for the network graphs.
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.0
 */
public class GenerateNetworkGraph {

	/**
	 * Instantiates a new GenerateNetworkGraph. Private to prevent instantiation.
	 * @version 1.0
	 */
	private GenerateNetworkGraph() {
		throw new IllegalStateException("Utility class");
	}


	/**
	 * Creates the Json necessary for the network graph.
	 * 
	 * @param fileName The source file containing the rows of failures
	 * @param graphTypeHeader The header in the file we want to examine,
	 * such as closestAudioFile or closestTranscriptionTestResultId
	 * @param largeReferences The references needed for a large sphere
	 * @param medReferences The references needed for a medium sphere
	 * @return The Json String for the network graph
	 * @since 1.0
	 */
	public static String generate(String fileName, String graphTypeHeader, int largeReferences, int medReferences) {
		String fileLine = "";
		int testResultIdIndex = -1;
		int closestTestResultIdIndex = -1;

		ArrayList<Integer> testResultIds = new ArrayList<>();
		ArrayList<String> connections = new ArrayList<>();

		//The SB for the graphical linkages between sphere
		StringBuilder data = new StringBuilder();

		//The SB for the sphere configs of each failure
		StringBuilder nodes = new StringBuilder();

		//Start looping thru the connection failures file
		try (BufferedReader br = new BufferedReader(new FileReader(fileName))) {
			while ((fileLine = br.readLine()) != null) {
				//Split the row into its separated columns
				String[] resultArr = fileLine.split("~", -1);

				//runDate means we're on the first row, so grab the header indices
				if (fileLine.startsWith("runDate")) {
					testResultIdIndex = Utils.getHeaderIndex(resultArr, "testResultId");
					closestTestResultIdIndex = Utils.getHeaderIndex(resultArr, graphTypeHeader);
				}
				else {
					//Grab the failure this row refers to, and its closest failure
					int testResultId = Integer.parseInt(resultArr[testResultIdIndex]);
					int closestTestResultId = getClosestTestResultId(resultArr, closestTestResultIdIndex, graphTypeHeader);


					//This should always be true
					if (!testResultIds.contains(testResultId)) {
						testResultIds.add(testResultId);
					}

					//Check if there's actually a closest match for this failure
					if (closestTestResultId > 0) {
						connections.add(testResultId + "~" + closestTestResultId);

						data.append("['" + testResultId + "', '" + closestTestResultId + "'],\r\n");
					}
				}
			}
		} catch (IOException e) {
			e.printStackTrace();
			System.exit(1);
		}


		//For each failed call, generate the Json for its sphere
		for (int x = 0; x < testResultIds.size(); x++) {
			nodes.append(createSphereJson(testResultIds.get(x), connections, largeReferences, medReferences));
		}

		return "<head>\r\n"
		+ "   <script src = \"jquery.2.1.3.min.js\"></script>\r\n"
		+ "   \r\n"
		+ "<script src=\"highcharts.js\"></script>\r\n"
		+ "<script src=\"networkgraph.js\"></script>\r\n"
		+ "\r\n"
		+ "</head>\r\n"
		+ "\r\n"
		+ "<body>\r\n"
		+ "<div id=\"container\" style=\"min-width: 320px;max-width: 800px;margin: 0 auto;height: 500px;margin-top: 50px;\"></div>\r\n"
		+ "\r\n"
		+ "<script>\r\n"
		+ "var dirDist" + largeReferences + " = \"#E8544E\",\r\n"
		+ "  dirDist" + medReferences + " = \"#FFD265\",\r\n"
		+ "  dirDistLess" + medReferences + " = \"#2AA775\";\r\n"
		+ "\r\n"
		+ "Highcharts.chart('container', {\r\n"
		+ "\r\n"
		+ "  chart: {\r\n"
		+ "    type: 'networkgraph',\r\n"
		+ "    marginTop: 80\r\n"
		+ "  },\r\n"
		+ "\r\n"
		+ "  title: {\r\n"
		+ "    text: 'Similar Step 1'\r\n"
		+ "  },\r\n"
		+ "\r\n"
		+ "  plotOptions: {\r\n"
		+ "    networkgraph: {\r\n"
		+ "      keys: ['from', 'to'],\r\n"
		+ "      layoutAlgorithm: {\r\n"
		+ "        enableSimulation: true,\r\n"
		+ "        integration: 'verlet',\r\n"
		+ "        linkLength: 100\r\n"
		+ "      }\r\n"
		+ "    }\r\n"
		+ "  },\r\n"
		+ "\r\n"
		+ "  series: [{\r\n"
		+ "    marker: {\r\n"
		+ "      radius: 13,\r\n"
		+ "    },\r\n"
		+ "    dataLabels: {\r\n"
		+ "      enabled: true,\r\n"
		+ "      linkFormat: '',\r\n"
		+ "      allowOverlap: true,style: {\r\n"
		+ "                    textOutline: false \r\n"
		+ "                }\r\n"
		+ "    },"
		+ "data: [\n"
		+ data.toString() + "\n"
		+ " ],\n"
		+ "   nodes: [\n"
		+ nodes.toString() + "\n"
		+ "]"
		+ "}]\r\n"
		+ "});\r\n"
		+ "</script>\r\n"
		+ "</body>";
	}


	/**
	 * Takes in a Step 1 Failed call row, as a String array,
	 * and finds the closest matching ID for the audio or
	 * the transcription, depending on what we pass in.
	 * @param resultArr The failed call row as a String arr
	 * @param closestTestResultIdIndex The index for which closest ID we want
	 * @param graphTypeHeader The type of closest ID we want (audio or transcription)
	 * @return The closest matching testResultId
	 * @since 1.0
	 */
	private static int getClosestTestResultId(String[] resultArr, int closestTestResultIdIndex, String graphTypeHeader) {
		int closestTestResultId = -1;

		//Check if we're looking for the audio file match, or the transcription match
		if (graphTypeHeader.equals("closestAudioFile")) {
			String closestAudioFile = resultArr[closestTestResultIdIndex];

			/*
			 * If the failed call row doesn't have .wav in the
			 * closestAudioFile, that means it didn't have one
			 */
			if (closestAudioFile.endsWith(".wav")) {
				closestTestResultId = Integer.parseInt(closestAudioFile.substring(0, closestAudioFile.indexOf("S")));
			}
		}
		else {
			closestTestResultId = Integer.parseInt(resultArr[closestTestResultIdIndex]);
		}

		return closestTestResultId;
	}


	/**
	 * Based on the number of references, generates the Json
	 * for the sphere in the network graph.
	 * @param testResultId The failed call's testResultId
	 * @param connections The list of calls and their respective closestTestResultId failure
	 * @param largeReferences The number of connections for a large-sized sphere
	 * @param medReferences The number of connections for a medium-sized sphere
	 * @return The Json String for this sphere
	 * @since 1.0
	 */
	private static String createSphereJson(int testResultId, ArrayList<String> connections, int largeReferences, int medReferences) {
		//For this specific testResultId, check how many children ultimately lead to it, and get the count
		ArrayList<Integer> previouslyCheckedIds = new ArrayList<>();
		previouslyCheckedIds.add(testResultId);

		/*
		 * Recursively get the number of references for this failed
		 * call. A reference is any hierarchal connec
		 */
		int references = getActualReferences(testResultId, connections, previouslyCheckedIds) + 1;

		if (references > 50) {
			references = 50;
		}

		String color = "";

		/*
		 * Based on how many hierarchal matches, references, there
		 * are, assign them the color clas used for their
		 * respective size.
		 */
		if (references >= largeReferences) {
			color = "dirDist" + largeReferences;
		}
		else if (references >= medReferences) {
			color = "dirDist" + medReferences;
		}
		else {
			color = "dirDistLess" + medReferences;
		}

		return "{\r\n" + 
		"        id: '" + testResultId + "',\r\n" + 
		"        marker: {\r\n" + 
		"          radius: " + references + ",\r\n" + 
		"        },\r\n" + 
		"        color: " + color + "\r\n" + 
		"      },\n";
	}


	/**
	 * Recursively grabs each matching failure to grab the true
	 * number of matches for a given parent failure.
	 * @param testResultId The failed call's testResultId
	 * @param connections The list of calls and their respective closestTestResultId failure
	 * @param previouslyCheckedIds The number of testResultIds we've already processed, prevent looping
	 * @return The number of references this parent failure has
	 * @since 1.0
	 */
	private static int getActualReferences(int testResultId, ArrayList<String> connections, ArrayList<Integer> previouslyCheckedIds) {
		int references = 0;

		for (int x = 0; x < connections.size(); x++) {
			String[] connection = connections.get(x).split("~");

			//A potential child (if its closest id matches)
			int checkedTestResultId = Integer.parseInt(connection[0]);

			//The closest test result id that, if matches, is being referenced
			int closestTestResultIdToCheck = Integer.parseInt(connection[1]);

			/*
			 * While looping thru all the connections, we found
			 * a failure whose closest match was our testResultId.
			 * Let's check if we haven't processed the match's
			 * owner, and process the owner.
			 */
			if (closestTestResultIdToCheck == testResultId) {
				references++;
				if (!previouslyCheckedIds.contains(checkedTestResultId)) {
					previouslyCheckedIds.add(checkedTestResultId);
					references += getActualReferences(checkedTestResultId, connections, previouslyCheckedIds);
				}
			}

		}

		return references;
	}
}
