package com.cyara.utilities;

import java.util.List;

/**
 * Various methods used throughout the samples for calculations.
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.0
 */
public class UtilsMath {

	/**
	 * Instantiates a new UtilsMath. Private to prevent instantiation.
	 * @version 1.0
	 */
	private UtilsMath() {
		throw new IllegalStateException("Utility class");
	}


	/**
	 *  Finds the mode of a given sorted Integer ArrayList
	 *  
	 *  @param responseTimes The sorted ArrayList containing 
	 *  the response times in milliseconds
	 *  @return The mode of the list
	 *  @since 1.0
	 */
	public static int findMode(List<Integer> responseTimes) {
		int maxValue = 0;
		int maxCount = 0;
		int trackedResponse = -1;
		int trackedCount = 0;

		for (int i = 0; i < responseTimes.size(); i++) {
			int resp = responseTimes.get(i);

			if (resp == trackedResponse) {
				trackedCount++;
			}
			else {
				if (trackedCount > maxCount) {
					maxCount = trackedCount;
					maxValue = trackedResponse;
				}

				trackedResponse = resp;
				trackedCount = 1;
			}
		}

		//Put this here so if the most common value is at the end of the array, we don't miss it
		if (trackedCount > maxCount) {
			maxValue = trackedResponse;
		}

		return maxValue;
	}


	/**
	 *  Finds the median of a given sorted Integer ArrayList
	 *  
	 *  @param responseTimes The sorted ArrayList containing 
	 *  the response times in milliseconds
	 *  @return The median of the list
	 *  @since 1.0
	 */
	public static int findMedian(List<Integer> responseTimes) {
		int middle = responseTimes.size() / 2;
		if (responseTimes.size() % 2 == 1) {
			return (int)responseTimes.get(middle);
		} else {
			return (int)((responseTimes.get(middle-1) + responseTimes.get(middle)) / 2.0);
		}
	}


	/**
	 *  Finds the mean of a given sorted Integer ArrayList
	 *  
	 *  @param responseTimes The sorted ArrayList containing 
	 *  the response times in milliseconds
	 *  @return The mean of the list
	 *  @since 1.0
	 */
	public static int findMean(List<Integer> responseTimes) {
		double sum = 0;
		for (int i = 0; i < responseTimes.size(); i++) {
			sum += responseTimes.get(i);
		}
		return (int)(sum / responseTimes.size());
	}


	/**
	 *  Finds the X percentile of a given sorted Integer ArrayList
	 *  
	 *  @param responseTimes The sorted ArrayList containing 
	 *  the response times in milliseconds
	 *  @param percentile The X percentile we want to grab
	 *  @return The mean of the list
	 *  @since 1.0
	 */
	public static int findPercentile(List<Integer> responseTimes, double percentile) {
		int index = (int) Math.ceil(percentile / 100.0 * responseTimes.size());
		return responseTimes.get(index-1);
	}


	/**
	 *  Finds the mode of a given sorted Integer ArrayList
	 *  
	 *  @param responseTimes The sorted ArrayList containing 
	 *  the response times in milliseconds
	 *  @return The mode of the list
	 *  @since 1.0
	 */
	public static double findModeDouble(List<Double> responseTimes) {
		double maxValue = 0;
		int maxCount = 0;
		double trackedResponse = -1;
		int trackedCount = 0;

		for (int i = 0; i < responseTimes.size(); i++) {
			double resp = responseTimes.get(i);

			if (resp == trackedResponse) {
				trackedCount++;
			}
			else {
				if (trackedCount > maxCount) {
					maxCount = trackedCount;
					maxValue = trackedResponse;
				}

				trackedResponse = resp;
				trackedCount = 1;
			}
		}

		//Put this here so if the most common value is at the end of the array, we don't miss it
		if (trackedCount > maxCount) {
			maxValue = trackedResponse;
		}

		return maxValue;
	}


	/**
	 *  Finds the median of a given sorted Integer ArrayList
	 *  
	 *  @param responseTimes The sorted ArrayList containing 
	 *  the response times in milliseconds
	 *  @return The median of the list
	 *  @since 1.0
	 */
	public static double findMedianDouble(List<Double> responseTimes) {
		int middle = responseTimes.size() / 2;
		if (responseTimes.size() % 2 == 1) {
			return responseTimes.get(middle);
		} else {
			return ((responseTimes.get(middle-1) + responseTimes.get(middle)) / 2.0);
		}
	}


	/**
	 *  Finds the mean of a given sorted Integer ArrayList
	 *  
	 *  @param responseTimes The sorted ArrayList containing 
	 *  the response times in milliseconds
	 *  @return The mean of the list
	 *  @since 1.0
	 */
	public static int findMeanDouble(List<Double> responseTimes) {
		double sum = 0;
		for (int i = 0; i < responseTimes.size(); i++) {
			sum += responseTimes.get(i);
		}
		return (int)(sum / responseTimes.size());
	}


	/**
	 *  Finds the X percentile of a given sorted Integer ArrayList
	 *  
	 *  @param responseTimes The sorted ArrayList containing 
	 *  the response times in milliseconds
	 *  @param percentile The X percentile we want to grab
	 *  @return The mean of the list
	 *  @since 1.0
	 */
	public static double findPercentileDouble(List<Double> responseTimes, double percentile) {
		int index = (int) Math.ceil(percentile / 100.0 * responseTimes.size());
		return responseTimes.get(index-1);
	}

	/**
	 * Calculates the percentage of divisor into the
	 * dividend. Cleanly converts that into a percentage
	 * with 2 decimal places.
	 * 
	 * @param dividend The number to divide into
	 * @param divisor The number that divides into the dividend
	 * @return The cleaned percentage
	 * @since 1.0
	 */
	public static double generateCleanPercentages(double dividend, double divisor) {
		double dirtyPct = dividend / divisor * 100.0;
		return Math.round(dirtyPct * 100.0) / 100.0;
	}


	/**
	 * Calculates the shortest character difference
	 * between two strings.
	 * 
	 * @param a The first string in the comparison
	 * @param b The second string in the comparison
	 * @return The character difference between the two strings
	 * @since 1.0
	 */
	public static int distance(String a, String b) {
		a = a.toLowerCase();
		b = b.toLowerCase();

		int [] costs = new int [b.length() + 1];
		for (int j = 0; j < costs.length; j++)
			costs[j] = j;
		for (int i = 1; i <= a.length(); i++) {
			costs[0] = i;
			int nw = i - 1;
			for (int j = 1; j <= b.length(); j++) {
				int cj = Math.min(1 + Math.min(costs[j], costs[j - 1]), a.charAt(i - 1) == b.charAt(j - 1) ? nw : nw + 1);
				nw = costs[j];
				costs[j] = cj;
			}
		}
		return costs[b.length()];
	}


	/**
	 * Randomly generates an integer within the range
	 * 
	 * @param min The smallest range of the random number
	 * @param max The largest range of the random number
	 * @return The random number between min and max
	 * @since 1.0
	 */
	public static int getRandomNumber(int min, int max) {
		return (int) ((Math.random() * (max - min)) + min);
	}

}
