package com.cyara.utilities;

import java.util.Calendar;
import java.util.Date;
import java.util.concurrent.TimeUnit;

/**
 * Various methods used throughout the samples for processing time.
 * <p>
 * This is not an official Cyara product and is not officially supported.
 * It is used for demonstration and training purposes. For assistance,
 * contact chris.ryan@cyara.com
 * 
 * @author Chris Ryan
 * @version 1.0
 */
public class UtilsTime {

	/**
	 * Instantiates a new UtilsTime. Private to prevent instantiation.
	 * @version 1.0
	 */
	private UtilsTime() {
		throw new IllegalStateException("Utility class");
	}


	/**
	 *  Convert the step response time from the API format
	 *  into the number of milliseconds that represents.
	 *  
	 *  @param responseTime The API format of a step response time
	 *  @return The response time in milliseconds
	 *  @since 1.0
	 */
	public static int getMillis(String responseTime) {
		//Split the API response time format
		if (!responseTime.contains(".")) {
			responseTime = responseTime + ".0000000";
		}
		String[] data = responseTime.split(":");

		//Parse each component of the format into its parts
		int hours = Integer.parseInt(data[0]);
		int minutes = Integer.parseInt(data[1]);
		int seconds = Integer.parseInt(data[2].substring(0, data[2].indexOf(".")));
		int millis = Integer.parseInt(data[2].substring(data[2].indexOf(".")+1, data[2].length())) / 10000;

		int time = seconds + 60 * minutes + 3600 * hours;

		long responseTimeMillis = TimeUnit.MILLISECONDS.convert(time, TimeUnit.SECONDS);

		responseTimeMillis += millis;

		return (int)responseTimeMillis;
	}


	/**
	 *  Converts a given date into time since C#-friendly epoch
	 *  
	 *  @param date The provided date
	 *  @return The C# time since epoch
	 *  @since 1.0
	 */
	public static long getUTCTicks(Date date) {
		Calendar calendar = Calendar.getInstance();
		calendar.setTime(date);

		//Add the ticks per millisecond and how many ticks ahead Epoch is from C# "epoch"
		return (calendar.getTimeInMillis() * 10000) + 621355968000000000L;
	}


	/**
	 *  Finds the first date of the quarter for 
	 *  a given date.
	 *  
	 *  @param date The provided date
	 *  @return The first date
	 *  @since 1.0
	 */
	public static Date getFirstDayOfThisQuarter(Date date) {
		Calendar cal = Calendar.getInstance();
		cal.setTime(date);
		cal.set(Calendar.DAY_OF_MONTH, 1);
		cal.set(Calendar.MONTH, cal.get(Calendar.MONTH)/3 * 3);
		return cal.getTime();
	}


	/**
	 *  Finds the last date of the quarter for 
	 *  a given date.
	 *  
	 *  @param date The provided date
	 *  @return The last date
	 *  @since 1.0
	 */
	public static Date getLastDayOfThisQuarter(Date date) {
		Calendar cal = Calendar.getInstance();
		cal.setTime(date);
		cal.set(Calendar.DAY_OF_MONTH, 1);
		cal.set(Calendar.MONTH, cal.get(Calendar.MONTH)/3 * 3 + 2);
		cal.set(Calendar.DAY_OF_MONTH, cal.getActualMaximum(Calendar.DAY_OF_MONTH));
		return cal.getTime();
	}


	/**
	 *  Finds the first date of the last quarter for 
	 *  a given date.
	 *  
	 *  @return The first date
	 *  @since 1.0
	 */
	public static Date getFirstDayOfLastQuarter() {
		Calendar cal = Calendar.getInstance();
		int year = cal.get(Calendar.YEAR);
		int month = cal.get(Calendar.MONTH);
		int presentQuarter = (month / 3) + 1;
		int lastQuarter = 0;

		if(presentQuarter - 1 > 0){
			lastQuarter = presentQuarter - 1;
		}else{
			lastQuarter = 4;
			year--;
		}

		month = lastQuarter * 3 - 1;

		cal.set(year, month, 1);
		return cal.getTime();
	}


	/**
	 *  Finds the last date of the last quarter for 
	 *  a given date.
	 *  
	 *  @return The first date
	 *  @since 1.0
	 */
	public static Date getLastDayOfLastQuarter() {
		Calendar cal = Calendar.getInstance();
		int year = cal.get(Calendar.YEAR);
		int month = cal.get(Calendar.MONTH);
		int presentQuarter = (month / 3) + 1;
		int lastQuarter = 0;

		if(presentQuarter - 1 > 0){
			lastQuarter = presentQuarter - 1;
		}else{
			lastQuarter = 4;
			year--;
		}

		month = lastQuarter * 3 - 1;

		int day = cal.getActualMaximum(Calendar.DAY_OF_MONTH);
		cal.set(year, month, day);

		return cal.getTime();
	}


	/**
	 *  Finds the first date of the year for a given date
	 *  
	 *  @param date The provided date
	 *  @return The first date
	 *  @since 1.0
	 */
	public static Date getFirstDayOfThisYear(Date date) {
		Calendar cal = Calendar.getInstance();
		cal.setTime(date);
		cal.set(Calendar.DAY_OF_MONTH, 1);
		cal.set(Calendar.MONTH, Calendar.JANUARY);
		return cal.getTime();
	}

}
